#!/usr/bin/python

# Copyright (C) 2002-2006
# Max-Planck-Institut fuer Radioastronomie Bonn
#
# Produced for the ALMA and APEX projects
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.  Correspondence concerning
# ALMA should be addressed as follows:
#
# Internet email: alma-sw-admin@nrao.edu
#
# Correspondence concerning APEX should be addressed as follows:
#
# "@(#) $Id: Entities.py,v 0.202 2006/05/05 14:40:53 vltsccm Exp $"
#
# Internet email: dmuders@mpifr-bonn.mpg.de

# The Python MBFITS Entity Objects library
#
# Who			        When		What
#
# C. Koenig, D.Paurat, MPIfR	03/20/2003	Initial version
#
# F. Schuller, MPIfR            05/13/2003      Initial version of operators
#                                               overloading
# F. Wyrowski, MPIfR            05/23/2003      keep number of dimensions in averageInt
# F. Wyrowski, MPIfR            06/12/2003      added averageChan
# E. Polehampton MPIfR          06/17/2003      update to averageInt to reset Integration number plus
#                                               some DATAPAR parameters.
# F. Schuller, MPIfR            06/27/2003      use the config.DEBUG global variable to control the
#                                               amount of debugging infos to be displayed, and
#                                               creation of a __str__ method to print the content
#                                               of an ObsEntity object
# E. Polehampton MPIfR          07/01/2003      update to plot and average methods for correct
#                                               shape of ARRAYDATA_DATA (add Feed number)
# E. Polehampton MPIfR          07/10/2003      update to averageChan method (for specified range of
#                                               channels)
# E. Polehampton MPIfR          07/14/2003      add cutIntegrations method
# F. Wyrowski, MPIfR            07/28/2003      add DOFIT options to plotContinuum and the gausslin
#						function used by the fit
# E. Polehampton MPIfR          08/01/2003      modify __sub__ to allow one argument to be a single float number.
#
# F. Schuller, MPIfR            08/12/2003      update __init__ and fill methods to the new way of reading XML file
#                                               moved all methods to the top level
# E. Polehampton MPIfR          08/13/2003      update for MBFITS 1.3
# F. Wyrowski, MPIfR            08/13/2003      add fit with SciPy in plotContinuum
# F. Schuller, MPIfR            08/13/2003      update fill method to handle the results of getObservation
# E. Polehampton MPIfR          08/20/2003      Update to __div__ method.
#                                               Print some parameters and return
#                                               offsets in Gauss fit.
# F. Schuller, MPIfR            08/21/2003      Update fill() method with last version of FitsWriter
# E. Polehampton MPIfR          08/22/2003      Update averaging methods to give correct integNum 
# E. Polehampton MPIfR          08/25/2003      Adjust loop in averageChannels
# E. Polehampton MPIfR          10/17/2003      Change inital values for Gauss fit in plotContinuum
# E. Polehampton MPIfR          10/20/2003      Correct INTGTIM definition in averageInt
#                                               Add focus option to plotContinuum
# E. Polehampton MPIfR          10/23/2003      Rearrange plotContinuum so that focus fit is a quadratic
# E. Polehampton MPIfR          10/30/2003      Update for MBFITS v1.5
# E. Polehampton MPIfR          11/03/2003      rename averageChan -> sumChannels and speed up method
#                                               update to averageInt method
# E. Polehampton MPIfR          11/06/2003      modification to select and cutIntegrations
#                                               add OTF plotting.
# E. Polehampton MPIfR          11/21/2003      Update for MBFITS v1.51 inc. order in ARRAYDATA_DATA
# E. Polehampton MPIfR          11/27/2003      Add option to make multiple pannels in plotContinuum()
# E. Polehampton MPIfR          12/04/2003      Allow plotting single point in _plot_spectrum()
# E. Polehampton MPIfR          12/09/2003      Remove excess dimension in FEBEPAR table in fill().
# E. Polehampton MPIfR          12/16/2003      Update averageInt() for case where number Ints. already 1
#                                               Add skydip option for plotting
# E. Polehampton MPIfR          12/18/2003      Add linear fit to skydip plotting
# E. Polehampton MPIfR          01/06/2004      in averageInt() adjust times for MBFITS 1.52 (remove MIDTIME)      
# E. Polehampton MPIfR          01/09/2004      Add option to make hardcopies in plot routines.
# E. Polehampton MPIfR          01/19/2004      Add error calculation in skydip fit.
# E. Polehampton MPIfR          02/09/2004      Change check in averageInt() for 2 phases
# E. Polehampton MPIfR          03/09/2004      Correct bug in averageInt(): (firstInt:lastInt+1)
#                                               Change label for Focus/Point plot to Counts (per second)
# E. Polehampton MPIfR          03/10/2004      averageInt() now calculates the sum and preserves integration time
# E. Polehampton MPIfR          03/24/2004      add wedge to OTF plot
# E. Polehampton MPIfR          04/16/2004      add a method to check the consistency of the channel scale
#                                               with another entity (I mean the frequency scale).
#                                               fix in creating attributes and filling because python does
#                                               not recognise attribute beginning with numbers (I add a '_')
# E. Polehampton MPIfR          04/28/2004      Overhaul overloaded operators: complete the _checkConsistent()
#                                               method with more checks and use this in __add__ etc.
#                                               Rename averageInt() to sumIntegrations()
# E. Polehampton MPIfR          04/29/2004      Change the way of supplying max/min values in
#                                               sumIntegrations(), sumChannels() and cutIntegrations()
# E. Polehampton MPIfR          04/30/2004      Import and use the Interpolator to carry out collapse of
#                                               the DATAPAR table for the sumIntegrations() method.
# E. Polehampton MPIfR          05/13/2004      Exponential fit for skydip in plotContinuum()
# E. Polehampton MPIfR          05/13/2004      Change from Numeric import * to import Numeric
# E. Polehampton MPIfR          05/18/2004      minor corrections in plotting
# E. Polehampton MPIfR          03/06/2004      change to numarray - but Numeric is still needed for plotting
# E. Polehampton MPIfR          07/06/2004      speed up select method. slight change to sumIntegr.
# E. Polehampton MPIfR          08/06/2004      Change to sumIntegrations to match Interpolator
# E. Polehampton MPIfR          14/06/2004      update plotContinuum to return all fit params for skydip
# E. Polehampton MPIfR          15/06/2004      print all results from pointing fit to screen
# E. Polehampton MPIfR          16/06/2004      first version of blanking in plotContinuum()
# E. Polehampton MPIfR          21/06/2004      in sumIntegrations make INTEGTIM array have a shape.
# E. Polehampton MPIfR          21/06/2004      update some doc strings.
# E. Polehampton MPIfR          28/06/2004      add method: removeAllBlanking()
# E. Polehampton MPIfR          28/06/2004      print FeBe/BB on plots.
#                                               change cutIntegrations to make much faster.
# E. Polehampton MPIfR          01/07/2004      Include linear method for skydip as separate option
# P. Schilke MPIfR              08/07/2004      Beautified pointing plots, added width
#                                               still missing: errors
# E. Polehampton MPIfR          11/07/2004      error and 2sigma clip for linear skydip
#                                               default to IntegNum=1 for plotSpectral
#                                               in sumIntegrations: remove integs with zero offset
#                                                 (caused by Interpolator when no AZEL monpoints)
#                                               move Feff calculation to plotContinuum. write value on plot
# E. Polehampton MPIfR          12/07/2004      Skydip exponential fit uses 2 paramters: tau, Feff
#                                               Tatm is fixed to Tamb. (old method retained as SKYDIP_old)
#                                               plot used points in blue, clipped points white for skydip
#                                               test in skydip processing to allow varying fixed params.
# E. Polehampton MPIfR          18/07/2004      adjust pointing plot labels.
# E. Polehampton MPIfR          03/08/2004      fix for plot limits if xmin=xmax or ymin=ymax
# E. Polehampton MPIfR          09/08/2004      update name monitor points attr to allow multiple points.
#                                               add logger
# E. Polehampton MPIfR          11/08/2004      add __reversePlanckCorrection(). Change name of Tchop -> Thot
#                                               all temps labeled with _RJ if rayleigh-jeans corrected
#                                               print Tatm for SKYDIP_old fit.
# E. Polehampton MPIfR          16/08/2004      Add blanking value as attribute.
# E. Polehampton MPIfR          17/08/2004      more statements to logger.
# E. Polehampton MPIfR          25/08/2004      changed skydip to fit TAsky. rather than counts.
#                                               (and ln(Tatm-TAsky) for lin)
# E. Polehampton MPIfR          01/09/2004      print absolute value of width
# E. Polehampton MPIfR          01/09/2004      return Ampl and wid from pointing plots
#                                               find plot limits by ignoring outside 10%  for onoffs.
# E. Polehampton MPIfR          21/09/2004      fix bug for ps plots: device now closed properly.
#                                               more print statements to logger.
# E. Polehampton MPIfR          24/09/2004      simplify skydip plot code. Add option to fix any parameter
#                                               implimented this to fix efficiency.
# E. Polehampton MPIfR          13/10/2004      add DATAPAR_LST in sumIntegrations- it should also be collapsed.
# E. Polehampton MPIfR          15/10/2004      Add Skydip fit using ATM model
#                                               Add possibility to calculate Chi2 and errors on fitted params
#                                               if errorBars are known.
#                                               Error bars cannot be plotted in ppgplot though..
#                                               add overplotting option to plotContinuum()
#                                               add fit description to skydip plots
#                                               remove self.plotNum as this was not working anyway.
# E. Polehampton MPIfR          26/10/2004      renamed the exp. skydip fit options and merged into one if statement.
# E. Polehampton MPIfR          02/11/2004      allow calculation of errors for pointings.
# E. Polehampton MPIfR          21/02/2005      calculate errors on pointings.
# E. Polehampton MPIfR          23/03/2005      update keywords for MBFITS v1.55.
#                                               set the default value for entity attributes to be self.blank
# E. Polehampton MPIfR          24/03/2005      update USEFEED dimension for MBFITS v1.55.
# E. Polehampton MPIfR          05/04/2005      split up the plotContinuum method: extract individual fit modes
#                                               rename _plot_spectrum() to _plotData()
#                                               check and correct the Feed indexing in plot methods.
#                                               add velocity axis to plotSpectral
#                                               add MBFTSVER to the obsEntity attributes
# E. Polehampton MPIfR          07/04/2005      find the baseband element and remove FEBE info for other bands
#                                               in the fill() method.
# E. Polehampton MPIfR          12/04/2005      small correction to skydip modes
# E. Polehampton MPIfR          13/04/2005      add plotAsOTF() method for offline case.
# E. Polehampton MPIfR          18/04/2005      allow more panels in plotContinuum. NumPlots->numPlots
# E. Polehampton MPIfR          19/04/2005      new method: __initialisePpgplotDevice()
# E. Polehampton MPIfR          22/04/2005      test possibility of using Boa to do plots.
#                                               in _plotData() - commented out for now.
# E. Polehampton MPIfR          29/04/2005      fix for plotContinuum when DOFIT but num points <2
# E. Polehampton MPIfR          02/05/2005      correct OTF plotting (should check SCANMODE not SCANTYPE)
#                                               correct plot labels. correct Focus labels.
# E. Polehampton MPIfR          08/05/2005      correct bug: skydip fitDescription was not being printed in plots.
# E. Polehampton MPIfR          11/05/2005      added ATMFreeTs to skydip fit modes.
# E. Polehampton MPIfR          13/05/2005      correct units for FOCUS fits. start to change labeling system
#                                               for plotting.
# E. Polehampton MPIfR          14/05/2005      catch exception if no data was in fits file in fill()
# E. Polehampton MPIfR          16/05/2005      correct labels for Trec, Tsys plots.
# E. Polehampton MPIfR          19/05/2005      get the monitor table header keywords as entity attributes
#                                               (needed to get USRFRAME)
# E. Polehampton MPIfR          21/05/2005      changed to use TEBB and BB-power in skydip fits.
# E. Polehampton MPIfR          27/05/2005      corrections to skydip fits.
# E. Polehampton MPIfR          09/06/2005      add skydip mode ATMfixed, print stddev of points about fit for skydips
# E. Polehampton MPIfR          14/06/2005      ignore underflow error messages during pointing fit optimisation
#                                               add y-axis labels for plots (from DATAUNIT attr).
# F. Schuller MPIfR             16/06/2005      for MONITOR table, read only Header in fill method
# E. Polehampton MPIfR          27/06/2005      - use add.areduce for INTEGTIM instead of add.reduce in sumIntegrations()
#                                               - set the font to be bigger for axis labels 
# E. Polehampton MPIfR          28/06/2005      - set the colour index for plots in __initialisePpgplotDevice()
#                                               - adjust the scaling of fonts for multiple plot panels
#                                               - increase the line width for fits
#                                               - remove some pieces of test code/commented out stuff
# E. Polehampton MPIfR          30/06/2005      save warning in errorStatus if averaging multiple
#                                               positions or phases.
# E. Polehampton MPIfR          11/07/2005      combine errorStatus in overloaded operators
# E. Polehampton MPIfR          16/07/2005      in _plotData(): ignore outside 10% in calculation of yrange
#                                               for CAL scantype (for Tsys and Trec plots)
# E. Polehampton MPIfR          16/07/2005      use velocity scale for OTF plots and improve colour table
# S. Leurini    MPIfR           05/08/2005      added blanking of MONITOR points
# E. Polehampton MPIfR          16/08/2005      - changed to Numeric
#                                               - add MONPOINT to select() and cutIntegrations()
# E. Polehampton MPIfR          14/09/2005      minor update to plotContinuum for yLabel & log statement
# E. Polehampton MPIfR          16/09/2005      add useCommandPosition() to recalculate LONGOFF and LATOFF
# E. Polehampton MPIfR          26/09/2005      fix for shorter name for Tsys/Trec plot
# A. Belloche MPIfR             20/10/2005      debug useCommandPosition(): convert the declination angle
#                                               from degree to radian to compute the cosine.
# Hafok                         02/11/2005      added subscanNum and Obstype in Header of DISPLAY
# Hafok                         06/12/2005      Changed ISWITCH string array to PHASE and PHASE1..N Keywords
#                                               for switched cycle observations.
# Hafok                         07/12/2005      added try..except in cutIntegrations to avoid problems
#                                               with non existant items in mbfits.
# Hafok                         14/12/2005      bugfix for detecting sum of different phases
# H.Hafok       , MPIfR 10/04/2006      changed order in ARRAY_DATA Integrations,Feeds,Channels
# Hafok                         05/05/2006      fix in PlotData to blank infinit values
""" This is the  MBFITS Entity Objects library.
(c) 2002-2006 MPIfR """


# Creates an Entity object from the given XML-structure

import XmlStructure
import string
import copy
import ppgplot
import mbfits  # needed to read the Tables in the fill method
import config
import numarray # needed for Error.setMode
import Numeric
#import Interpolator # needed for sumIntegrations()
import BoaMessageHandler as CalibratorLog
#from scipy.optimize import leastsq  # for fitting
# this is for the plotting routines
#from Bogli import Plot

# This module is to be used only with MB-FITS files
# Thus, store the corresponding XML description in a global variable
XmlStruc = XmlStructure.XmlStructure('MBFITS')

class ObsEntity:

    # Creator function of the ObsEntity class
    def __init__(self):

        """
        create an empty Entity object with defaults set to
        the Python None object
        """

        global XmlStruc

        # blanking value
        self.blank = -999.0

        # Get the XML tables representing one observation
        SCANMBFITS      = XmlStruc.XmlScan.tables['SCAN-MBFITS']
        FEBEPARMBFITS   = XmlStruc.XmlScan.tables['FEBEPAR-MBFITS']
        ARRAYDATAMBFITS = XmlStruc.XmlScan.tables['ARRAYDATA-MBFITS']
        DATAPARMBFITS   = XmlStruc.XmlScan.tables['DATAPAR-MBFITS']
        MONITORMBFITS   = XmlStruc.XmlScan.tables['MONITOR-MBFITS']

        # Create attributes for the ObsEntity object 
        self.getXMLHeaderKeywords(SCANMBFITS)
        self.getXMLKeywords(FEBEPARMBFITS)
        self.getXMLKeywords(ARRAYDATAMBFITS)
        self.getXMLKeywords(DATAPARMBFITS)
        self.getXMLHeaderKeywords(MONITORMBFITS)


        # get logger 
        self.Logger = CalibratorLog.Logger()

        # Physical constants
        self.h = 6.626176e-34
        self.k = 1.380662e-23
        self.c = 299792458.0
 
        
        # don't need this anymore..i think.
        # self.plotNum  = 0
        

    #####################################################
    # Definition of functions to extract the informations
    # from the XmlStruc object. These functions are used
    # by the __init__ method of the ObsEntity class.
    
    def createEntityAtribute(self,XmlElement, TableName=None):
        """
        >> createEntityAtribute(XmlElement, TableName=None)
        
        creates attributes from either a header or a binary table
        and sets them to the Python None object
        """
    
        if XmlElement.hasChildNodes() == 1:
            # search for the value of the tag (nodeType=3:Text) containing
            # its keyword-name. Somewhat ugly, but I didn't find a method
            # for direct access to the enclosed text
            for XmlNode in XmlElement.childNodes:
                if XmlNode.nodeType == 3:
                    name = str(XmlNode.data)
        else:
            raise EntitiesError, "Keyword not defined"
        # Takes "-" out of the name, if given 
        if '-' in name:
            names = string.split(name,'-')
            name = string.join(names,'_')
    
        # Add the TableName to the attribute's name
        if TableName == None:
            # names beginning with numbers are not valid attributes in python
            # add a _ before 
            if name[0] in ['1','2']:
                name = "_"+name
            # initialise the attribute to the blanking value
            setattr(self,name,self.blank)
        else:
            # TableName without '-MBFITS'
            TableName = TableName[:-7] 
            name = string.join(string.split(TableName,'-'),'_')+'_'+name
            setattr(self,name,self.blank)
    
    
    def getXMLHeaderKeywords(self,Table):
        """
        >> getXMLHeaderKeywords(Table)
        
        creates keywords from a XML table's header keywords  """
        
        TblHeader = Table.getElementsByTagName('HEADER')[0]
        TblHeaderKeywds = TblHeader.childNodes
        for Keyword in TblHeaderKeywds:
            if Keyword.nodeType == 1:
                self.createEntityAtribute(Keyword)        
      
    def getXMLBinTblKeywords(self,Table):
        """
        >> getXMLBinTblKeywords(Table)

        creates keywords from a XML table's binary table columns """
        
        BinTable = Table.getElementsByTagName('TABLE')[0]
        BinTableColumns = BinTable.childNodes
        for Keyword in BinTableColumns:
            if Keyword.nodeType == 1:
                self.createEntityAtribute(Keyword, Table.tagName)
    
    def getXMLKeywords(self,Table):
        """
        >> getXMLKeywords(Table)
        
        creates keywords from a XML table's header keywords
        and binary table columns  """
        
        self.getXMLHeaderKeywords(Table)
        self.getXMLBinTblKeywords(Table)
            
    # End of XML related functions definitions
    ##########################################
    


    def fill(self,fitsFile,TableList):
        """
        >> fill(fitsFile,TableList)
        
        fills the ObsEntity object with data given from a list of tables
        corresponding to an observation, from the MBFits file pointed to by fptr
        """

        global XmlStruc

        # TableList is the result of the getObservation method, and contains
        # the observing mode (element [0]) and the list of HDU numbers

        # read the MBFITS version of the file from the Primary header
        primary = mbfits.Table(Parent=fitsFile)
        primary.HduNum = 1
        setattr(self,'MBFTSVER',primary.Header.readKeyword('MBFTSVER')[1:-1])
        
        # to avoid doing 2 loops (and thus reading the tables twice), fill
        # attributes corresponding to both the keywords and binary tables
        # for all tables except PRIMARY
        self.ColNames = []
        for i in xrange(1,len(TableList)):

            Table = mbfits.Table(Parent=fitsFile)  # instantiate (and initialise) a new Table object
            name = TableList[i][1]
            if config.DEBUG:
		print "Now reading table",name,"with HDU number =",TableList[i][0]
            if name == 'MONITOR-MBFITS':   # for MONITOR tables, read only header
                Table.readHeader(TableList[i][0])
            else:
                Table.read(TableList[i][0])

            # Get the HeaderKeywords and set them as a new attributes
            # to the ObsEntity object
	    if config.DEBUG:
	        print
	        print "ObsEntity.fill: setting new attributes from the header keywords..."

            for Key in Table.Header.KeywordDict:
                Value = Table.Header.KeywordDict[Key]
                if '-' in Key:
                    Keys = string.split(Key,'-')
                    Key = string.join(Keys,'_')
                if Key[0] in ['1','2']:
                    Key = "_"+Key
                if hasattr(self,Key):
                    setattr(self,Key,Value[0])
                    if config.DEBUG > 1:
                        print "...",Key

 

            if name != 'MONITOR-MBFITS':
                # Reads out the column names from the Xml-Structure
                if config.DEBUG:
                    print
                    print "ObsEntity.fill: setting new attributes from the contents of binary tables..."

                namesub= string.join(string.split(name,'-'),'_')[:-7]
                BinTable = XmlStruc.XmlScan.tables[name].getElementsByTagName('TABLE')[0]
                for Element in BinTable.childNodes:
                    for node in Element.childNodes:
                        if node.nodeType == 3:
                            self.ColNames.append(node.data)
                # Get the BinaryTableColumns and set them as a new attribute to the ObsEntity object
                for ColName in self.ColNames:
                    AttrName = namesub+'_'+string.join(string.split(ColName,'-'),'_')
                    if hasattr(self, AttrName):
                        try:
                            # There is only ever one row in the FEBEPAR table
                            # (each FEBE has its own FEBEPAR table)
                            # Therefore the extra dimension giving the row number has no
                            # physical significance...so only fill the first row
                            # sometimes the extra dimension doesn't seem to exist though
                            # (for USEBAND and REFFEED).
                            if Table.Name == 'FEBEPAR-MBFITS':
                                try:
                                    setattr(self,AttrName,Table.BinTable.Data[ColName][0])
                                except:
                                    setattr(self,AttrName,Table.BinTable.Data[ColName])
                                # WARNING! - remove this - it is a dirty fix for Effelsberg files!!!!
                                # elif Table.Name == 'ARRAYDATA-MBFITS' and ColName == 'DATA':
                                #     setattr(self,AttrName,Table.BinTable.Data[ColName][:,:,Numeric.NewAxis])
                            else:
                                # FIXME: The scheme fails for files with a single integration
                                #        in ARRAYDATA/DATAPAR. One should map this into a Numeric
                                #        array rather than a scalar, but then many calculations
                                #        using "array[-1]" still fail as Numeric does not support
                                #        this in that case.
                                #        This should really be fixed in mbfits.py, but since it
                                #        this will affect many users, it first needs to be
                                #        checked if everyone would be OK with such a change.
                                item = Table.BinTable.Data[ColName]
                                setattr(self,AttrName,item)

                            if config.DEBUG > 1:
                                print 'ObsEntity.fill: '+Table.Name,
                                print "...",AttrName
                        except:
                            # this means that an MBFITS keyword in the xml description was
                            # not present in the FITS file.
                            if config.DEBUG > 1:
                                print 'ObsEntity.fill: '+Table.Name,
                                print "...",AttrName,"- not present in FITS file.."
                            pass

        # sort out the FEBEPAR table information
        # the USEFEED array is dimensioned per baseband. The obsEntity is only for a single
        # ARRAYDATA so its better to remove the extra information for the
        # other basebands at this point.
        febeAttrSingleDim=['USEBAND','NUSEFEED','BOLCALFC','BOLREFGN','GAINELE1','GAINELE2']
        febeAttrDoubleDim=['USEFEED','BESECTS','APEREFF','BEAMEFF','ETAFSS','HPBW','ANTGAIN',\
                           'TCAL','FLATFIEL','BOLDCOFF','GAINIMAG']
        mbFits1_54exceptions = ['USEFEED','FEEDTYPE','NUSEFEED','BOLDCOFF','TCAL','BESECTS']
        try:
            # FS: in the next few lines, I have added int() around self.BASEBAND,
            # self.NUSEBAND and self.NUSEFEED because it was of type 'string' on
            # some computers/systems (Mac OS X, but also some recent Linux Suse 10)
            bandElement = self.FEBEPAR_USEBAND.tolist().index(int(self.BASEBAND))
            for attrName in vars(self).keys():
                if attrName[0:7] == 'FEBEPAR':
                    oldValue = getattr(self,attrName)
                    # fix to allow MBFITS 1.54 to continue working:
                    if self.MBFTSVER.count('1.54') and attrName[8:] in mbFits1_54exceptions:
                        pass
                    # otherwise for MBFITS v1.55 (and the other keys in 1.54):
                    elif attrName[8:] in febeAttrSingleDim:
                        setattr(self,attrName,oldValue[bandElement])
                    elif attrName[8:] in febeAttrDoubleDim:
                        setattr(self,attrName,oldValue[bandElement,:])
                    elif attrName[8:] == 'FEEDTYPE':
                        feedType   = oldValue
                        lenPerBand = len(feedType)/int(self.NUSEBAND)
                        start      = bandElement*lenPerBand
                        setattr(self,attrName,feedType[start:start+int(self.NUSEFEED)])                
        except Exception,error:
            # this means that the fits file was not filled correctly
            if hasattr(error,'msg'):
                self.Logger.logger.logWarning(str(error.msg))
                raise EntitiesError('Could not get the data for that subscan - check the FITS file was written OK')
            else:
                self.Logger.logger.logWarning(str(error.args))
                raise EntitiesError('Could not get the data for that subscan - check the FITS file was written OK')


        # check that there was data in the ARRAYDATA table. DATAPAR should always exist if
        # there is an ARRAYDATA and would be filled with blanking values if there was a problem
        if len(self.ARRAYDATA_DATA) == 0:
            raise EntitiesError('There was no data in the ARRAYDATA table - check the FITS file!')

 
	if config.DEBUG:
            print
            print "...Done filling :)"
            print

	if config.DEBUG > 1:
            print self


    # PLOT ROUTINES **** these use NUMERIC ******
    # because ppgplot is currently compiled using Numeric


    def __initialisePpgplotDevice(self,numPlots,psName):
        """ This initialises the ppgplot device
            either xwindow or postscript. """
        
        # set the ppgplot device
        if psName == 'none':
            device = '/xs'
        else:
            # but why are postscript files created always black and white??!
            device = str(psName)+'/ps'
            self.Logger.logger.logInfo("Plotting to a postscript file in current directory with name: %s"%(psName))
 
        try:
            # this command closes any devices that may already be open.
            ppgplot.pgbeg(device)
        except:
            raise EntitiesError("Failed trying to open PPGPLOT device in ppgplot.pgbeg: %s"%(device))
        
        # set up the number of plot panels
        if numPlots == 2:
            ppgplot.pgsubp(1,2) #<- 2 panels
            if config.DEBUG:
                self.Logger.logger.logInfo("Plotting with 2 panels")
            ppgplot.numPlots = 2
        elif numPlots in [3,4]:
            ppgplot.pgsubp(-2,2) #<- 4 panels
            if config.DEBUG:
                self.Logger.logger.logInfo("Plotting with 4 panels")            
            ppgplot.numPlots = 4
        elif numPlots in [5,6]:
            ppgplot.pgsubp(-3,2) #<- 6 panels
            if config.DEBUG:
                self.Logger.logger.logInfo("Plotting with 6 panels")            
            ppgplot.numPlots = 6
        elif numPlots in [7,8]:
            ppgplot.pgsubp(-2,4) #<- 8 panels
            if config.DEBUG:
                self.Logger.logger.logInfo("Plotting with 8 panels")            
            ppgplot.numPlots = 8
        elif numPlots > 8:
            ppgplot.pgsubp(-3,4) #<- 12 panels
            if config.DEBUG:
                self.Logger.logger.logInfo("Plotting with 12 panels")            
            ppgplot.numPlots = 12
        else:
            ppgplot.numPlots = 1

        # the line width
        lineWidths = [8, 6, 4, 4, 4, 4, 4, 4, 2, 2, 1, 1]
        ppgplot.lineWidth = lineWidths[ppgplot.numPlots-1]

        # ** Determine the font size and position for the fit labels **
        # the font size automatically scales but the scaling depends on the number of
        # plot panels in the vertical direction.
        # therefore may only need to modify for 4 plots as the others scale ok.
        fontSizes   = [2, 2, 1.5, 1.5, 1.5, 1.5, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0]
        fontFactors = [1.0, 0.7, 1.0, 1.0, 1.0, 1.0, 0.7, 0.7, 0.7, 0.7, 0.7, 0.7]
        ppgplot.fontSize   = fontSizes[ppgplot.numPlots-1]
        ppgplot.fontFactor = fontFactors[ppgplot.numPlots-1]
        
        # decide on the colour to use as default for fits here
        # red = 2, green = 3, blue = 4
        if not hasattr(ppgplot,'colourIndex'):
            ppgplot.colourIndex = 3

        # FIXME! set the background colour here too if possible.
        

    def plotSpectral(self,integNum=1,feedNum=1,numPlots=1,renew=1,psName='none'):
        """
        >> plotSpectral(integNum=1,feedNum=1,numPlots=1,renew=1,psName='none')
        
        Plot spectral line data vs. channels to the /xs or /ps devices
        using ppgplot
        Input:
              integNum = Integration number (first integration = 1: default)
              feedNum  = Feed number (first Feed is 1: default)
              numPlots : Divide the page into this number of plots (default is 1)
                         can be 1, 2, 4, 6, 8, 12
                      
              renew: logical: if 1 then start a new page (<- default)
                            if 0 then plot next plot on current page
                            (needed when multiple plots per page)

              psName: name of postscript file: if present hardcopy is produced
        """

        # convert input Integnum and Feednum to the element number:
        integElement = integNum - 1
        # the feed number is the physical feed number. We must determine its element in the USEFEED array.
        feedElement = self.FEBEPAR_USEFEED.tolist().index(feedNum)

        # initialise the device (or add to ps file if already open)
        if renew:
            self.__initialisePpgplotDevice(numPlots,psName)
        elif ppgplot.pgqinf('HARDCOPY') == 'YES':
            self.Logger.logger.logInfo("Adding to the postscript file with name: "+ppgplot.pgqinf('FILE'))


        # ------- OTF -------
        if self.SCANMODE == 'OTF':
            xLabel = 'Velocity (km/s)'
            yLabel = 'Integration number'
            self._plotData(0,0,feedElement,xLabel,yLabel,histo='OTF')

            if config.DEBUG > 1:
                self.Logger.logger.logInfo("Plotting an OTF...")

        # ------- special case for ATM model results -------
        elif self.SCANTYPE == 'ATMSPECTRUM':
            #integNum = Numeric.array(integNum,'i')
            channels = Numeric.array(self.ARRAYDATA_DATA[0,feedElement,:])
            spectrum = Numeric.array(self.ARRAYDATA_DATA[integNum-1,feedElement,:])
            xLabel=''
            yLabel=''
            self._plotData(channels,spectrum,feedElement,xLabel,yLabel)
            
            # oplot other integrations?
            #if Numeric.size(integElement) > 1:
            #    colour=2
            #    for integ in integElement[1:]:
            #        spectrum = Numeric.array(self.ARRAYDATA_DATA[integ,feedElement,:])
            #        ppgplot.pgsci(colour)
            #        ppgplot.pgbin(channels,spectrum)
            #        colour = colour + 1

            #ppgplot.pgsci(1)

        # ------- normal spectra -------
        elif self.OBSTYPE == 'TREC' or self.OBSTYPE == 'TSYS':
            spectralAxis = Numeric.array(xrange(int(self.CHANNELS)))
            xLabel = "Channels"
            spectrum = Numeric.array(self.ARRAYDATA_DATA[integElement,feedElement,:])
            yLabel = "%s (K)"%(self.OBSTYPE)
            self._plotData(spectralAxis,spectrum,feedElement,xLabel,yLabel)
        else:
            if config.DEBUG:
                self.Logger.logger.logInfo("Plotting integration number %i of %i"%(integNum,len(self.ARRAYDATA_DATA)))
                self.Logger.logger.logInfo("Feed %i of %i"%(feedNum,self.NUSEFEED))

            try:
                # get the x-axis in velocity   
                spectralAxis = Numeric.zeros(self.CHANNELS,'d')
                for i in xrange(self.CHANNELS):
                    spectralAxis[i] = self._1CRVL2R + ((i+1-self._1CRPX2R)*self._11CD2R)
                xLabel = "Velocity (km/s)"
            except:
                # just use channel scale
                spectralAxis = Numeric.array(xrange(int(self.CHANNELS)))
                xLabel = "Channels"

            spectrum = Numeric.array(self.ARRAYDATA_DATA[integElement,feedElement,:])
            # get the units for the y-axis from self.DATAUNIT
            yLabel = getattr(self, 'DATAUNIT', '')
            self._plotData(spectralAxis,spectrum,feedElement,xLabel,yLabel)

        # close the graphics device: 
        if (numPlots == 1):
            if config.DEBUG:
                self.Logger.logger.logInfo("Closing PPGPLOT plotting device")
            ppgplot.pgclos()
        else:
            if config.DEBUG:
                self.Logger.logger.logInfo("Expecting more plots: PPGPLOT Device left open.")


    def plotContinuum(self,feedNum=1,dofit='NO',numPlots=1,renew=1,oplotFit=0,psName='none'):
        """
        >> plotContinuum(feedNum=1,dofit='NO',numPlots=1,renew=1,oplotFit=0,psName='none')
        
        Plotting of continuum data.

        optional Input:

            Feed number (feed number counting from 1)  default=1

            dofit: The string, 'DOFIT' performs a fit. Anything else does not.
                   The fit performed depends on the SCANTYPE:
                              FOCUS:  Quadratic
                              POINT:  Gaussian + baseline
                              SKYDIP: default is a 3 param exponential fit.
                                      (see also below)
                                                     
            numPlots: Divide the page into this number of plots (default is 1)
                      can be 1, 2, 4, 6, 8, 12
                      
            renew: logical: if 1 then start a new page (<- default)
                            if 0 then plot next plot on current page
                            (needed when multiple plots per page)

            oplotFit: integer: if greater than 0 then the next call to plot will
                      put the fit only over the top of previous plot (if used with renew=0
                      AND if the first plot had numPlots set greater than 1 -> 3 is a good value to use
                      unless you want to divide the page up)

            psName: name of postscript file: if present then hardcopy is produced

            The plot device is closed after a single call if numPlots=1.
            ...therefore if you want to add another plot to the
               current page, numPlots should not be 1 until the last one you want to add.
               For postscript output the device MUST be closed => set numPlots=1 in last
               call for multiple plots.

            The fitting methods all require that some processing of the data has been done
            in advance so it is better to reduce the data first using reduce() and then plot
            using plotPointing()
                  plotFocus()
                  plotSkydip()

        """

        # the feed number is the physical feed number. We must determine its element in the USEFEED array.
        feedElement = self.FEBEPAR_USEFEED.tolist().index(feedNum)

        # initialise the device (or add to ps file if already open)
        if renew:
            self.__initialisePpgplotDevice(numPlots,psName)
        elif ppgplot.pgqinf('HARDCOPY') == 'YES':
            self.Logger.logger.logInfo("Adding to the postscript file with name: "+ppgplot.pgqinf('FILE'))

        # get the Y-axis points
        try:
            spectrum = Numeric.array(self.ARRAYDATA_DATA[:,feedElement,0])
        except:
            spectrum = Numeric.array(1)
        yLabel = getattr(self, 'DATAUNIT', 'Counts')

        # error bars? used for calculating chi2 in fits. But cannot plot them with ppgplot!!
        try:
            errorBars = Numeric.array(self.ErrorBars[:,feedElement])
        except:
            errorBars = Numeric.array(1)
        
        # Get the correct X-axis points
        # plotType = 0 means dots, = 1 means histogram
        xLabel = 'Integration Number'
        try:
            if (self.SCANDIR == 'ALAT'):
                channels = Numeric.array(self.DATAPAR_LATOFF[:,]*3600.0)
                xLabel = 'Offset (arcseconds)'
                driftType = 'ELEVATION'
                plotType  = 1
            elif (self.SCANDIR == 'ALON'):
                channels = Numeric.array(self.DATAPAR_LONGOFF[:,]*3600.0)
                xLabel = 'Offset (arcseconds)'
                driftType = 'AZIMUTH'
                plotType  = 1
            elif (self.SCANTYPE[:5] == 'FOCUS'):
                if self.SCANTYPE[6:] in ('XTILT','YTILT'):
                    # Units deg -> change to arcseconds
                    channels = Numeric.array(self.FOCUSPOS[:,])*3600.0
                    xLabel = 'Offset (arcseconds)'
                else:
                    channels = Numeric.array(self.FOCUSPOS[:,])
                    xLabel = 'Offset (mm)'
                driftType = self.SCANTYPE
                plotType  = 0          
            elif (self.SCANTYPE[0:6] == 'SKYDIP'):
                channels = Numeric.array(self.AIRMASS[:,])
                xLabel = 'Airmass'
                driftType = 'SKYDIP'
                plotType  = 0                    
            else:
                channels = Numeric.array(range(len(self.ARRAYDATA_DATA[:,feedElement,0])))
                self.Logger.logger.logWarning("Could not determine the scan type so am using an integration number scale only.")
                driftType = ''
                plotType=1
        except:
            channels = Numeric.array(1)
            driftType = ''
            plotType=1

        # now remove values that were blanked
        # if a value in spectrum is equal to blankingValue then it will be removed before plotting and fitting
        # but not irrevocably removed from the data outside of this method.
        if spectrum != 1 and channels !=1:
            originalSpectrum = spectrum            
            spectrum  = Numeric.compress(Numeric.not_equal(originalSpectrum,self.blank),spectrum)
            channels  = Numeric.compress(Numeric.not_equal(originalSpectrum,self.blank),channels)
            if errorBars != 1:
                errorBars = Numeric.compress(Numeric.not_equal(originalSpectrum,self.blank),errorBars)

        # plot the data
        if not oplotFit:
            self._plotData(channels,spectrum,feedElement,xLabel,yLabel,errorBars,histo=plotType)

        # ******** FITTING ****************************************************
        offset      = 0.0, 0.0, 0.0, 0.0
        offsetError = 0.0, 0.0, 0.0, 0.0
        # only works if more than 2 points in the array
        if dofit == 'DOFIT' and len(spectrum) > 2:

            # set colour for the fit and fit result labels
            # (default value set to red when opening ppgplot device)
            ppgplot.pgsci(ppgplot.colourIndex)
            # unless overplotting in which case use the value of oplotFit to set the colour
            if oplotFit:
                ppgplot.pgsci(oplotFit)
                
            # increase the thickness of the line
            ppgplot.pgslw(ppgplot.lineWidth)
                                        
            label          = 'NO FIT'                       
            label2         = ""
            label3         = ""
            fitDescription = ""
            
            try:
                x_range = max(channels)-min(channels)
                x_array = min(channels)+x_range*(Numeric.array(xrange(100),'d')/100.0)

                # ------- FOCUS -------
                if self.SCANTYPE[:5] == 'FOCUS':
                    offset, label = self.__fitFocus(channels, spectrum, x_range, x_array)

                # ------- SKYDIP -------
                elif self.SCANTYPE == 'SKYDIP_LIN':                   
                    offset, offsetError, label, label2, fitDescription = \
                            self.__fitSkydipLin(feedElement, channels, spectrum, x_range, x_array)
                    driftType = 'SKYDIP_LIN'
            
                elif (self.SCANTYPE[0:10] == 'SKYDIP_ATM'):
                    offset, offsetError, label, label2, label3, fitDescription = \
                            self.__fitSkydipATM(feedElement, channels, spectrum, x_range, x_array, errorBars)

                elif self.SCANTYPE[0:10] == 'SKYDIP_EXP':
                    offset, offsetError, label, label2, label3, fitDescription = \
                            self.__fitSkydipExp(feedElement, channels, spectrum, x_range, x_array, errorBars)

                # ------- POINTING -------
                elif self.SCANTYPE == 'POINT':
                    offset, offsetError, label, label2, label3 = self.__fitGauss(channels, spectrum, x_range, x_array)

                else:
                    self.Logger.logger.logWarning("Cannot fit scan type %s"%(self.SCANTYPE))
            
            # if no fit at all could be performed
            except TypeError,name:
                self.Logger.logger.logWarning("A TypeError has occurred: %s"%(name.args[0]))
            except ValueError,name:
                msg = "A ValueError has occurred! usually related to Numeric/Numeric conversion: %s"%(name.args[0])
                self.Logger.logger.logWarning(msg)

            # line width back to normal
            ppgplot.pgslw(1)

            # set the font size for the labels
            ppgplot.pgsch(ppgplot.fontSize)

            if not oplotFit:
                if driftType == 'SKYDIP':
                    ppgplot.pgmtxt('B',-6*ppgplot.fontFactor,0.5,0.0,label)
                    ppgplot.pgmtxt('B',-4*ppgplot.fontFactor,0.5,0.0,label2)
                    ppgplot.pgmtxt('B',-2*ppgplot.fontFactor,0.5,0.0,label3)
                else:
                    ppgplot.pgmtxt('B',-14*ppgplot.fontFactor,0.6,0.0,label)
                    ppgplot.pgmtxt('B',-12*ppgplot.fontFactor,0.6,0.0,label2)
                    ppgplot.pgmtxt('B',-10*ppgplot.fontFactor,0.6,0.0,label3)
              
                ppgplot.pgmtxt('B',-14*ppgplot.fontFactor,0.1,0.0,driftType)
                # back to normal size for skydip fit description
                ppgplot.pgsch(1)
                if driftType == 'SKYDIP':
                    # ppgplot.pgmtxt('B',-26,0.1,0.0,fitDescription)
                    ppgplot.pgmtxt('B',-15*ppgplot.fontFactor,0.5,0.0,fitDescription)

            # reset colour and font size to be default (white)
            ppgplot.pgsci(1)
            ppgplot.pgsch(1)
                    
        # ** end of FITTING ************
        
        # close the graphics device: 
        if (numPlots == 1):
            if config.DEBUG:
                self.Logger.logger.logInfo("Closing PPGPLOT plotting device")
            ppgplot.pgclos()
        else:
            if config.DEBUG:
                self.Logger.logger.logInfo("Expecting more plots: PPGPLOT Device left open.")

        # only return results if a fit was performed
        if dofit == 'DOFIT':
            return offset, offsetError

           
    # *** INDIVIDUAL FITTING METHODS ****************************

    def __fitFocus(self, channels, spectrum, x_range, x_array):
        """ Quadratic fit for FOCUS """

        # initials: [Y-offset, width parameter, X-offset] 
        if x_range > 0:
            initials = [max(spectrum),4*(max(spectrum)-min(spectrum))/(x_range**2),channels[Numeric.argmax(spectrum)]]
        else:
            initials = [max(spectrum),1.0,channels[Numeric.argmax(spectrum)]]

        if config.DEBUG:
            self.Logger.logger.logInfo("Initial values: "+str(initials))
        
        plsq   = leastsq(self.__quad_residuals, initials, args=(spectrum, channels))           
        ppgplot.pgline(x_array,self.__quadratic(plsq[0],x_array))
        
        if config.DEBUG:
            print "Fitted parameters: "
            print "   Y-offset: ", plsq[0][0]
            print "   Width parameter: ", plsq[0][1]
            print "   Focus Offset: ", plsq[0][2]

        if self.SCANTYPE[6:] in ('XTILT','YTILT'):
            unit = 'arcsec'
        else:
            unit = 'mm'
                                           
        label  = "Offset = %4.3f %s" %(plsq[0][2],unit)

        return plsq[0][2], label


    def __fitSkydipLin(self, feedElement, channels, spectrum, x_range, x_array):
        """ Linear Skydip fit """
        
        initials=[0.0,0.4]
        fitDescription = "Linear fit to ln(P(Tamb)-P(TAsky))"

        if config.DEBUG:
            print "Initial values: ",initials
                    
        plsq0 = leastsq(self.__lin_residuals, initials, args=(spectrum, channels))
        dataMinusModel = self.__lin_residuals(plsq0[0],spectrum,channels)
 
        # ** sigma clipping **
        # perform a 2 stage sigma clipping before arriving at the final fit.
        # set the sigma clip level:
        clipLevel = 2.0
                        
        # first stage to remove far outliers
        stdDev = Numeric.sqrt((Numeric.add.reduce(dataMinusModel**2))/(len(spectrum)-1.0))

        newSpectrum = Numeric.compress(Numeric.less(abs(dataMinusModel),clipLevel*stdDev),spectrum)
        newChannels = Numeric.compress(Numeric.less(abs(dataMinusModel),clipLevel*stdDev),channels)

        initials = [plsq0[0][0],plsq0[0][1]]
        plsq1 = leastsq(self.__lin_residuals, initials, args=(newSpectrum,newChannels))
        dataMinusModel = self.__lin_residuals(plsq1[0],newSpectrum,newChannels)
                    
        # second stage to refine the clipping:        
        stdDev = Numeric.sqrt((Numeric.add.reduce(dataMinusModel**2))/(len(newSpectrum)-1.0))
            
        finalSpectrum = Numeric.compress(Numeric.less(abs(dataMinusModel),clipLevel*stdDev),newSpectrum)
        finalChannels = Numeric.compress(Numeric.less(abs(dataMinusModel),clipLevel*stdDev),newChannels)

        self.Logger.logger.logInfo("%i points clipped"%(len(spectrum)-len(finalSpectrum)))
        
        initials = [plsq1[0][0],plsq1[0][1]]
        plsq = leastsq(self.__lin_residuals, initials, args=(finalSpectrum,finalChannels))
        
        forwardEff = Numeric.exp(plsq[0][0])/self.Bamb[feedElement]
        tau        = -plsq[0][1]
        Tatm       = 0

        # plot the fit
        ppgplot.pgline(x_array,self.__linear(plsq[0],x_array))
        # plot the points used:
        # change the colour to blue here to show points used in the fit
        ppgplot.pgsci(4)
        ppgplot.pgpt(channels,spectrum,-31)
        # set colour back to default
        ppgplot.pgsci(ppgplot.colourIndex)

        if config.DEBUG > 1:
            print "Fitted parameters: "
            print "   intercept: ", plsq[0][0]
            print "   gradient: ", plsq[0][1]

        # ** error **
        try:
            dataMinusModel = self.__lin_residuals(plsq[0],finalSpectrum,finalChannels)
            # error on a least squares minimisation:
            pointError = (1./(len(spectrum)-2.0))*Numeric.add.reduce(dataMinusModel**2)
            intError   = pointError/len(finalSpectrum)
            xDash      = finalChannels - (Numeric.add.reduce(finalChannels)/len(finalChannels))
            xTmp       = (xDash**2)/pointError
            gradError  = 1/(Numeric.add.reduce(xTmp))
        except:
            gradError  = 0            

        offset = forwardEff, tau, Tatm, plsq[0][0]
        offsetError = 0, Numeric.sqrt(gradError), 0, 0

        label  = "tau = %4.3f \(2233) %4.3f"%(offset[1],offsetError[1])
        label2 = "Feff = %4.3f"%(offset[0])

        return offset, offsetError, label, label2, fitDescription
                        

    def __fitSkydipATM(self, feedElement, channels, spectrum, x_range, x_array, errorBars=1.0):
        """ ATM fit for Skydip """
                        
        # this uses the ATM routines to fit the skydip
        # there is no clip at the moment.
        
        from Calibrator import ATMmodel  # for skydip fitting
                        
        # Initialise model and make global parameters
        # - these are needed by the fit method
        global ATM, OpacityResults
        ATM = ATMmodel(self.measuredTamb[0],\
                       self.Pamb[0],\
                       self.SITEELEV/1e3,\
                       self.humidity[0],\
                       self.LapseRate[0],\
                       self.ScaleHeight[0])  
        OpacityResults = ATM.getSkyOpacity(self,feedElement)

        # Estimate the frequency to invert the black body equation as the combination
        # of signal and image side bands.
        freq = (self.Freq_Signal[0] + self.GainImage[0]*self.Freq_Image[0])/(1.0+self.GainImage[0])

        # organise free and fixed parameters.
        # The last 4 numbers give the index where param is located in order:
        # PWV, Feff, Tatm, Tspill
        if self.SCANTYPE[10:] == '3FREE':
            initials = [0.5,0.95,self.calcBlackBodyPower(270,freq)]
            fixedParam = [self.Bamb[feedElement], 0, 1, 3, 2]            
            fitDescription = "Fit using ATM: 3 free parameters"
        elif self.SCANTYPE[10:] == 'FIXTS':
            initials = [0.5,0.95]
            fixedParam = [self.Bamb[feedElement],self.Bspill[feedElement], 0, 1, 2, 3]            
            fitDescription = "Fit using ATM: Tspill fixed, Feff, PWV free"
        elif self.SCANTYPE[10:] == 'FIXWATER':
            initials = [0.95]
            fixedParam = [self.Water[feedElement],self.Bamb[feedElement],self.Bspill[feedElement], 1, 0, 2, 3]            
            fitDescription = "Fit using ATM: Tspill and PWV fixed, Feff free"
        elif self.SCANTYPE[10:] == 'FIXEFF':
            initials = [1.0]
            fixedParam = [self.Feff[feedElement],self.Bamb[feedElement],self.Bspill[feedElement], 0, 1, 2, 3]            
            fitDescription = "Fit using ATM: Tspill and Feff fixed, PWV free"
        elif self.SCANTYPE[10:] == 'FREETS':
            initials = [0.5,self.calcBlackBodyPower(270,freq)]
            fixedParam = [self.Feff[feedElement],self.Bamb[feedElement], 0, 2, 3, 1]            
            fitDescription = "Fit using ATM: Tspill, PWV free; Feff fixed"
        elif self.SCANTYPE[10:] == 'FIXED':
            initials = []
            fixedParam = [self.Water[feedElement],self.Bspill[feedElement],self.Feff[feedElement],self.Bamb[feedElement], 0, 2, 3, 1]            
            fitDescription = "Fit using ATM: all fixed"
        else:
            # unknown fit method
            return (0,0,0,0), (0,0,0,0), '', '', '', ''

        if config.DEBUG:
            print "initials",initials

        # The fit
        if len(initials):
            plsq = leastsq(self.__ATMFit_residuals, initials, args=(spectrum, channels,fixedParam,errorBars))
        else:
            plsq = Numeric.array([[]],'d')
      
        # Plot the final fit
        if type(plsq[0]) == float:
            results = Numeric.array([plsq[0]],'d')
        else:
            results = plsq[0]            
        finalParams = results.tolist()+fixedParam

        ppgplot.pgline(x_array,self.__ATMFit(finalParams,x_array))

        # Estimate the goodness of fit using std dev of residuals
        try:   
            finalModel  = self.__ATMFit(finalParams,channels)
            resid       = self.__ATMFit_residuals(plsq[0],spectrum,channels,fixedParam,errorBars)
            resid_final = Numeric.add.reduce((resid/finalModel)**2)
            pstddev     = 100.0*Numeric.sqrt(resid_final/(len(resid)-1.))
            self.Logger.logger.logInfo("Stddev of points about best fit = %.3f percent"%(pstddev))
        except:
            self.Logger.logger.logInfo("Could not calculate stddev of points about best fit")
 
                        
        waterError  = ""
        FeffError = ""
        # #####################################################################
        errors = 1
        # this is for calculating error bars
        # can only happen if errorBars are present and if there are more than 2 fixed params.
        if errors and len(errorBars) > 1 and len(results) > 1:
            ParamError = self.calcErrors(results,spectrum,channels,fixedParam,errorBars,"ATMFit")
            waterError  = " \(2233) %4.3f"%(ParamError[fixedParam[-4]])
            FeffError = " \(2233) %4.3f"%(ParamError[fixedParam[-3]])
        else:
            # ParamError always has 4 elements.
            ParamError = Numeric.zeros(4,'d')
        TsError = ParamError[fixedParam[-1]]
        # #######################################################################################        
                        
        water      = finalParams[fixedParam[-4]]
        forwardEff = finalParams[fixedParam[-3]]
        Tatm       = 0  # Tatm not used in the ATM fit.
        Tspill     = self.calcTEBB(finalParams[fixedParam[-1]],freq)
                        
        # calculate tau from best fit PWV
        tau_s, tau_i = ATM.getFinalOpacities(water,OpacityResults)
         
        if self.SCANTYPE[10:] == 'FIXEFF' or self.SCANTYPE[10:] == 'FREETS':
            FeffError = '(fixed)'
        if self.SCANTYPE[10:] == 'FIXWATER':
            waterError = '(fixed)'
        
        label  = "tau_s = %4.3f"%(tau_s)
        label2 = "PWV = %4.3f%s mm"%(water,waterError)
        label3 = "Feff = %4.3f%s"%(forwardEff,FeffError)

        offset      = forwardEff, tau_s, Tatm, Tspill
        offsetError = ParamError[fixedParam[-3]],ParamError[fixedParam[-4]],0.0,TsError

        return offset, offsetError, label, label2, label3, fitDescription

        
    def __fitSkydipExp(self, feedElement, channels, spectrum, x_range, x_array, errorBars=1.0):
        """ Exponential fit for Skydip """
        

        # Estimate the frequency to invert the black body equation as the combination
        # of signal and image side bands.
        if self.SCANTYPE[10:] != 'SIMPLE':
            freq = (self.Freq_Signal[0] + self.GainImage[0]*self.Freq_Image[0])/(1.0+self.GainImage[0])
        else:
            freq = 0.0

        # This fits power using skydip formula
        # Can cope with fixing different params
        # The last 4 entries in fixedParam give the index where each param is found in order
        # tau, Feff, Tatm, Tspill.

        if self.SCANTYPE[10:] == '3FREE':
            initials = [0.5, 1.0, self.calcBlackBodyPower(spectrum[-1],freq)]
            fixedParam = [self.Bspill[feedElement], 0, 1, 2, 3]
            fitDescription = "Exp fit with 3 free parameters and Tspill fixed"
        elif self.SCANTYPE[10:] == '2FREE':
            initials = [1.0,1.0]
            fixedParam = [self.Bamb[feedElement],self.Bspill[feedElement], 0, 1, 2, 3]            
            fitDescription = "Exp fit with Tatm=Tamb and Tspill fixed"
        elif self.SCANTYPE[10:] == 'FIXEFF':
            initials = [0.5, self.calcBlackBodyPower(spectrum[-1],freq)]
            fixedParam = [self.Feff[feedElement], self.Bspill[feedElement], 0, 2, 1, 3]
            fitDescription = "Exp fit with Feff and Tspill fixed."
        else:
            # otherwise use the SIMPLE method
            initials = [0.5, max(spectrum), 1.0]
            fixedParam = [1.0, 0, 1, 3, 2]  
            fitDescription = "Completely free exponential fit. No fixed constraints."
                  
        # first fit:
        plsq0 = leastsq(self.__skydip_residuals, initials, args=(spectrum, channels,fixedParam,errorBars))
        dataMinusModel = self.__skydip_residuals(plsq0[0],spectrum,channels,fixedParam,1.0)
        
        # ** sigma clipping **
        # perform a 2 stage sigma clipping before arriving at the final fit.
        clipLevel = 2.0
                        
        # first stage to remove far outliers:
        stdDev = Numeric.sqrt((Numeric.add.reduce(dataMinusModel**2))/(len(spectrum)-1.0))
        
        newSpectrum = Numeric.compress(Numeric.less(abs(dataMinusModel),clipLevel*stdDev),spectrum)
        newChannels = Numeric.compress(Numeric.less(abs(dataMinusModel),clipLevel*stdDev),channels)

        initials = plsq0[0]
        plsq1 = leastsq(self.__skydip_residuals, initials, args=(newSpectrum,newChannels,fixedParam,errorBars))
        dataMinusModel = self.__skydip_residuals(plsq1[0],newSpectrum,newChannels,fixedParam,1.0)
        
        # second stage to refine the clipping:
        stdDev = Numeric.sqrt((Numeric.add.reduce(dataMinusModel**2))/(len(newSpectrum)-1.0))
        
        finalSpectrum = Numeric.compress(Numeric.less(abs(dataMinusModel),clipLevel*stdDev),newSpectrum)
        finalChannels = Numeric.compress(Numeric.less(abs(dataMinusModel),clipLevel*stdDev),newChannels)
        
        self.Logger.logger.logInfo("%i points clipped"%(len(spectrum)-len(finalSpectrum)))
        
        initials = plsq1[0]
        plsq = leastsq(self.__skydip_residuals, initials, args=(finalSpectrum,finalChannels,fixedParam,errorBars))
        
        # plot the final fit
        finalParams = plsq[0].tolist()+fixedParam
        ppgplot.pgline(x_array,self.__skydipfit(finalParams,x_array))

        # calculate the std dev of points about the fit
        try:   
            finalModel  = self.__skydipfit(finalParams,finalChannels)
            resid       = self.__skydip_residuals(plsq[0],finalSpectrum,finalChannels,fixedParam,errorBars)
            resid_final = Numeric.add.reduce((resid/finalModel)**2)
            pstddev     = 100.0*Numeric.sqrt(resid_final/(len(resid)-1.))
            self.Logger.logger.logInfo("Stddev of points about best fit = %.3f percent"%(pstddev))
        except:
            self.Logger.logger.logInfo("Could not calculate stddev of points about best fit")
        
        # plot the points used:
        ppgplot.pgsci(4)
        ppgplot.pgpt(finalChannels,finalSpectrum,-31)
        ppgplot.pgsci(ppgplot.colourIndex)
        
        results = plsq[0] 
        
        tauError   = ""
        FeffError  = ""
        TaErrorStr = ""
        # ####################################################################
        errors = 1
        # this is for calculating error bars
        # can only happen if errorBars are present and if there are more than 2 fixed params.
        if errors and len(errorBars) > 1 and len(results) > 1:
            ParamError = self.calcErrors(results,spectrum,channels,fixedParam,errorBars,"skydip")
            tauError  = " \(2233) %4.3f"%(ParamError[fixedParam[-4]])
            FeffError = " \(2233) %4.3f"%(ParamError[fixedParam[-3]])
        else:
            # ParamError always has 4 elements.
            ParamError = Numeric.zeros(4,'d')
        TaError = ParamError[fixedParam[-2]]
        TsError = ParamError[fixedParam[-1]]
        # #######################################################################################        

        # add in zeros for fixed parameters
        results = results.tolist()
        while len(results) < 4:
            results.append(0)

        # for the simple fit, only tau can be determined.
        if self.SCANTYPE[10:] == 'SIMPLE':
            results[1:4] = [0.0,0.0,0.0]

        # model best fit is in Power => transform back to T_EBB
        tau        = results[fixedParam[-4]]
        forwardEff = results[fixedParam[-3]]
        Tatm       = self.calcTEBB(results[fixedParam[-2]],freq)
        Tspill     = self.calcTEBB(finalParams[fixedParam[-1]],freq)

        TaError = TaError
        if TaError:
            TaErrorStr = " \(2233) %4.3f"%(TaError)
            
        if not forwardEff and self.SCANTYPE[10:] != 'SIMPLE':
            forwardEff = self.Feff[feedElement]
            FeffError  = " (fixed)"

        label  = "tau = %4.3f%s"%(tau,tauError)
        label2 = ""
        label3 = ""

        if forwardEff:
            label2 = "Feff = %4.3f%s"%(forwardEff,FeffError)
        if Tatm:
            label3 = "Tatm = %4.3f%s"%(Tatm,TaErrorStr)

        offset      = forwardEff, tau, Tatm, Tspill
        offsetError = ParamError[fixedParam[-3]],ParamError[fixedParam[-4]],TaError, 0.0

        return offset, offsetError, label, label2, label3, fitDescription


    def __fitGauss(self, channels, spectrum, x_range, x_array):
        """ Gaussian fit for Pointing """
        
        # GAUSSIAN FIT
        # Initial values for fit: norm.; centre; width; intercept; gradient
        initials = [max(spectrum)-min(spectrum),channels[Numeric.argmax(spectrum)],x_range/3.0,spectrum[0], 0]
        # no fixed param - this is just the order of free params (which has to be
        # appended to the end of the list of free params in __gauss_residuals())
        fixedParam = [0,1,2,3,4]
        
        # the fit:
        plsq = leastsq(self.__gauss_residuals, initials, args=(spectrum, channels, fixedParam, 1.0))
        finalParams = plsq[0].tolist()+fixedParam
        # plot on top of the data:
        ppgplot.pgline(x_array,self.__gauss_lin(finalParams,x_array))
        
        intercept = plsq[0][3]
        gradient  = plsq[0][4]
        width     = abs(plsq[0][2])
        corr      = plsq[0][1]
        amplitude = plsq[0][0]
        
        corError = ""
        widError = ""
        ampError = ""
        # #####################################################################
        errors = 1
        
        # The residuals in the best fit:
        residuals = self.__gauss_residuals(plsq[0],spectrum, channels,fixedParam,1.0)
        sqdiff    = 0.0
        npoints   = 0.0             
        for k in xrange(len(residuals)):
            # Calculate the RMS in the baseline 2FWHM away from the centre of the
            # fitted Gaussian
            if channels[k] < (corr - 2.0*width) or channels[k] > (corr + 2.0*width):
                sqdiff  = sqdiff + residuals[k]**2
                npoints += 1.0
        # don't bother with errors if it was a rubbish fit - if fit corr is outside the
        # x-scale range.
        if npoints > 1.0 and sqdiff >= 0.0 and corr > min(channels) and corr < max(channels):
            stdDev = Numeric.sqrt(sqdiff/(npoints-1.0))
            errorBars = Numeric.ones(len(channels),'d')*stdDev
            self.Logger.logger.logInfo("Baseline RMS over range %4.0f to %4.0f and %4.0f to %4.0f = %4.3f"\
                                       %(min(channels),corr - 2.0*width,corr + 2.0*width,max(channels),stdDev))            
        else:
            errorBars = Numeric.array(1)
                                        
        # Error calculation can only happen if errorBars are present.
        if errors and len(errorBars) > 1:
            results  = plsq[0][0:3]
            fixedParam = [plsq[0][3],plsq[0][4],0,1,2,3,4]
            ParamError = self.calcErrors(results,spectrum,channels,fixedParam,errorBars,"gauss")
            ampError = " \(2233) %4.3f"%(abs(ParamError[fixedParam[-5]]))
            corError = " \(2233) %4.3f"%(abs(ParamError[fixedParam[-4]]))
            widError = " \(2233) %4.3f"%(abs(ParamError[fixedParam[-3]]))
        else:
            # ParamError always has 4 elements.
            ParamError = Numeric.zeros(4,'d')
                        
        # ###############################################################################

        if config.DEBUG:
            print "Fitted parameters: "
            print "   Baseline intercept: ", intercept
            print "   Baseline gradient: ", gradient
            print "   Gaussian width: ", width
            print "   Gaussian offset: ", corr
            print "   Gaussian normalisation: ", amplitude

                
        offset = corr, amplitude, width
        offsetError = ParamError[fixedParam[-4]], ParamError[fixedParam[-5]], ParamError[fixedParam[-3]]
                    
        label  = "Off = %5.2f%s"%(corr,corError)
        label3 = "Wid = %3.1f%s"%(width,widError)
        if amplitude < 1.0:
            label2 = "Amp = %4.3f%s"%(amplitude,ampError)
        elif amplitude < 1e5:
            label2 = "Amp = %5.1f%s"%(amplitude,ampError)
        else:
            label2 = "Amp = %4.1e%s"%(amplitude,ampError)

        return offset, offsetError, label, label2, label3



    # =============================================================================
        
    def calcErrors(self,fitResults,spectrum,channels,fixedParam,errorBars,fitMeth):
        """
        >> calcErrors(fitResults,spectrum,channels,fixedParam,errorBars,fitMeth) -> ParamError

        Calculates errors on the fitted parameters for skydips if error bars are
        provided for each data point.
        This is done via Chi squared.
        Each free parameter in tern is varied about its best value. The other params are
        reoptimised at each position.
        DeltaChi2 is examined and the error on that parameter is given when DeltaChi2
        equals 1
        """
        # initialise the parameter errors
        ParamError  = Numeric.zeros(4,'d')
        # model parameters (for which errors will be found) + fixed parameters
        finalParams = fitResults.tolist()+fixedParam       

        # calculate the chi2 at best fit
        if fitMeth == "ATMFit":
            tmpChiSqu = (self.__ATMFit_residuals(fitResults,spectrum,channels,fixedParam,errorBars))**2
        elif fitMeth == "skydip":
            tmpChiSqu = (self.__skydip_residuals(fitResults,spectrum,channels,fixedParam,errorBars))**2
        elif fitMeth == "gauss":
            tmpChiSqu = (self.__gauss_residuals(fitResults,spectrum,channels,fixedParam,errorBars))**2
        else:
            return ParamError
        
        if fitMeth == "gauss":
            Nparam=5
        else:
            Nparam=4
        
        minimumChiSquared = Numeric.add.reduce(tmpChiSqu)
        #print "Min chi squared = ",minimumChiSquared

        # Vary each parameter in turn about its best fit value,
        # re-fitting the others at every point.
        # First, fix all params at their best values
        newFixedParams = finalParams
        # loop over free parameters
        for pInx in xrange(len(fitResults)):
            #print "Parameter ",pInx+1,"of",len(fitResults)
            bestValue = [fitResults[pInx]]
            # put this value in the fixed params
            newFixedParam = bestValue+fixedParam
            newResults = []
            paramIndex = newFixedParam[-Nparam:]
            # alter the indices which give the order of the parameters in fixedParam:
            for pI in xrange(Nparam):
                indx = paramIndex[-Nparam+pI]
                if indx == pInx:
                    newFixedParam[-Nparam+pI] = len(fitResults)-1
                elif indx > pInx and indx < len(fitResults):
                    newFixedParam[-Nparam+pI] = paramIndex[-Nparam+pI]-1                                
                if indx != pInx and indx < len(fitResults):
                    newResults.append(fitResults[indx])                                    
            # the newResult has 1 less degree of freedom
            # therefore deltaChiSq is chi2 distributed with 1 degr of freedom
            # find where deltaChi2 is 1.0, re-fitting the remaining params.
            err=0.1  # start with 10% error
            deltaChiSquared=0.0
            #print fitResults,fixedParam
            #print newResults,newFixedParam
            niter   = 0
            while (1.0-deltaChiSquared)**2 > 0.01**2:
                newFixedParam[0] = (1.0+err)*bestValue[0]
                if fitMeth == "ATMFit":
                    plsq = leastsq(self.__ATMFit_residuals,newResults,args=(spectrum,channels,newFixedParam,errorBars))
                    if type(plsq[0]) == float: loopres = Numeric.array([plsq[0]],'d')
                    else: loopres = plsq[0]
                    tmpChiSqu = (self.__ATMFit_residuals(loopres,spectrum,channels,newFixedParam,errorBars))**2
                    deltaChiSquared = Numeric.add.reduce(tmpChiSqu) - minimumChiSquared
                elif fitMeth == "skydip":
                    plsq = leastsq(self.__skydip_residuals,newResults,args=(spectrum,channels,newFixedParam,errorBars))
                    if type(plsq[0]) == float: loopres = Numeric.array([plsq[0]],'d')
                    else: loopres = plsq[0]
                    tmpChiSqu = (self.__skydip_residuals(loopres,spectrum,channels,newFixedParam,errorBars))**2
                    deltaChiSquared = Numeric.add.reduce(tmpChiSqu) - minimumChiSquared
                elif fitMeth == "gauss":
                    plsq = leastsq(self.__gauss_residuals,newResults,args=(spectrum,channels,newFixedParam,errorBars))
                    if type(plsq[0]) == float: loopres = Numeric.array([plsq[0]],'d')
                    else: loopres = plsq[0]
                    tmpChiSqu = (self.__gauss_residuals(loopres,spectrum,channels,newFixedParam,errorBars))**2
                    deltaChiSquared = Numeric.add.reduce(tmpChiSqu) - minimumChiSquared
                else:
                    self.Logger.logger.logInfo("Could not calculate errors on fit parameters: unknown fit method")
                    return ParamError
                    
                #print "deltaChiSquared",deltaChiSquared, ", error = "+str(err*100.0)+"%"
                finalErr = err
                err = err/Numeric.sqrt(abs(deltaChiSquared))
                niter += 1
                # stop after 10 iterations in defeat or if deltaChiSquared was < 0
                if niter == 10 or deltaChiSquared < 0:
                    finalErr = 0.0
                    deltaChiSquared = 1.0
                    
                
            ParamError[pInx] = finalErr*bestValue[0]
            # print "best value = "+str(bestValue[0])+"+/-"+str(finalErr*bestValue[0])
            # print " ________________"

        return ParamError


    # ------------------------------------------------------
    #
    #  Gauss + linear baseline function needed for fitting
    #
    def __gauss_lin(self,parameters,x):
        """
        >> __gauss_lin(parameters,x) -> (a*exp(-4*log(2)*((x-b)/c)**2)+d+e*x)
        parameters = a,b,c,d,e  
        Gauss + linear baseline function needed for fitting"""

        # ignore underflow errors that occur during the optimisation
        # (these occur during bad fits and ignoring them just means
        #  that the warning messages are not printed to the screen)
        # still need to use numarray here because its not possible in Numeric
        numarray.Error.setMode(underflow="ignore")
                
        a = parameters[parameters[-5]]
        b = parameters[parameters[-4]]
        c = parameters[parameters[-3]]
        d = parameters[parameters[-2]]
        e = parameters[parameters[-1]]
        # a,b,c,d,e = parameters

        # the Numeric exponential function doesn't work for
        # high negative values in the exponent -> clip at -6e2. 1e399=infinity
        result = Numeric.array(a*Numeric.exp(Numeric.clip(-4*Numeric.log(2)*((x-b)/c)**2,-6e2,1e399))+d+e*x)

        numarray.Error.setMode(underflow='warn')

        return result

    def __gauss_residuals(self,p, y, x,fp,errorBars):
        """
        >> __gauss_residuals(p, y, x) -> residuals
        used for scipy least square fitting"""
        # join fitparams to fixedparams        
        parameters = p.tolist()+fp

        # set errorBars to 1.0 to get just residuals
        return (y - self.__gauss_lin(parameters,x))/errorBars

    def __quadratic(self,parameters,x):
        """
        >> __quadratic(parameters,x) -> a - b*((x-c)**2)
        Quadratic needed for fitting""" 
        a,b,c = parameters
        return a - b*((x-c)**2)

    def __quad_residuals(self,p, y, x):
        """
        >> __quad_residuals(p, y, x) -> residuals
        used for scipy least square fitting"""
        return y - self.__quadratic(p,x)

    def __linear(self,parameters,x):
        """
        >> __linear(parameters,x) -> a + b*x
        Straight line fit"""
        a,b = parameters
        return a + b*x

    def __lin_residuals(self,p, y, x):
        """
        >> __lin_residuals(p, y, x) -> residuals
        used for scipy least square fitting
        """
        return y - self.__linear(p,x)



    def __skydipfit(self,parameters,x):
        """
        >> __skydipfit(self,parameters,x) ->
        (1-Feff)*Ts + Feff*(1-Numeric.exp(-tau*numx))*Ta)
        exponential fit for skydip"""

        #(the last 4 entries in parameters give the index of each param.
        Ba = parameters[parameters[-2]]    # power
        Bs = parameters[parameters[-1]]    # power
        Feff = parameters[parameters[-3]]  
        tau = parameters[parameters[-4]]

        fitBskyA = (1-Feff)*Bs + Feff*(1-Numeric.exp(Numeric.clip(-tau*x,-6e2,1e399)))*Ba
        
        return fitBskyA


    def __skydip_residuals(self,p, y, x, fp,errorBars):
        """
        >> __skydip_residuals(p, y, x, fp) -> residuals
        used for scipy least square fitting"""
        # join fitparams to fixedparams        
        parameters = p.tolist()+fp
        return (y - self.__skydipfit(parameters,x))/errorBars

    def __ATMFit(self,parameters,x):

        # Ta = parameters[parameters[-2]] not used.
        Bs    = parameters[parameters[-1]]
        Feff  = parameters[parameters[-3]]
        water = parameters[parameters[-4]]  # replace tau for water here.

        fitBskyA      = Numeric.zeros(len(x),'d')
        for i in xrange(len(x)):
            # OpacityResults and skydipFeed are global! - they're set up in the skydip plot methods.
            Bsignal, Bimage = ATM.getBsky(water, x[i], OpacityResults)
            BskyModel       = (Bsignal + Bimage*self.GainImage[0]) / (1.0 + self.GainImage[0])         
           
            fitBskyA[i]     = (1.0-Feff)*Bs + Feff*BskyModel

        return fitBskyA

    def __ATMFit_residuals(self,p, y, x, fp,errorBars):
        """
        """
        # join fitparams to fixedparams
        parameters = p.tolist()+fp
        # if error bars have not been specified then errorBars has length 1 and value 1.0
        return (y - self.__ATMFit(parameters,x))/errorBars
        


    def plotAsOTF(self,feedNum,psName='none'):
        """
        >> plotAsOTF(feedNum, psName='none')

        Plot any multi-integration spectral object as a 2D spectrum.
        
        """
        # plot an observation as 2D spectrum but only if multiple integrations
        if len(self.ARRAYDATA_DATA) > 1 and self.CHANNELS > 1:
            feedElement = self.FEBEPAR_USEFEED.tolist().index(feedNum)
            # set the ppgplot device
            if psName == 'none':
                device = '/xs'
            else:
                device = str(psName)+'/ps'
                self.Logger.logger.logInfo("Plotting to a postscript file in current directory with name: "+psName)
            try:   
                ppgplot.pgbeg(device)
            except:
                errorMessage = 'plotAsOTF(): failed trying to open device in ppgplot.pgbeg:'+device
                raise EntitiesError(errorMessage)

            xLabel = 'Velocity (km/s)'
            yLabel = 'Integration Number'
            self._plotData(0,0,feedElement,xLabel,yLabel,histo='OTF')

            # close the graphics device: currently only for postscript plots
            if device == str(psName)+'/ps':
                if config.DEBUG:
                    self.Logger.logger.logInfo("Closing Postscript device")
                ppgplot.pgclos()            
        # otherwise just do it as a normal spectrum
        elif self.CHANNELS > 1:
            self.Logger.logger.logInfo("Only one integration: plotting as a normal spectrum")
            self.plotSpectral(feedNum=feedNum,psName='none')
        else:
            self.Logger.logger.logInfo("Only one integration: plotting as continuum")
            self.plotContinuum(feedNum=feedNum,psName='none')
           
   

     # --------------------------------------------------------

    def _plotData(self,channels,spectrum,feedElement,xLabel,yLabel,errorBars=[1],histo=1):
        """
        >> _plotData(channels,spectrum,feedElement,errorBars=[1],histo=1)
        
        called by plotContinuum() and plotSpectral()
        uses the program ppgplot to draw a simple graph
        Careful: the feed is specified by element number in the USEFEED array here
        ErrorBars: currently does nothing: the error bar plotting routines are not in ppgplot!!
    
        """

        # -------------------------------------------------------------------
        # plotting using BOA - can only be commented in when BOA available in Chile
        ##Graphic = BogliBusiness.BogliB()
        #title = 'Scan: '+str(self.SCANNUM)+'  Feed: '+str(self.FEBEPAR_USEFEED[feedElement])+ \
        #        '   Source: '+str(self.OBJECT)+'  FeBe: '+self.FEBE+'  BaseBand: '+str(self.BASEBAND)
        #if histo == 1 and len(channels) > 1:
        #    Plot.plot(channels,spectrum,labelX=xLabel,labelY=yLabel,caption=title,style='b')
        #elif histo == 'OTF':
        #    A = self.ARRAYDATA_DATA[:,feedElement,:]
        #    Plot.draw(A,labelX=xLabel,labelY=yLabel,caption=title,wedge=1)
        #else:
        #    # this plots both single and multiple points
        #    Plot.plot(channels,spectrum,labelX=xLabel,labelY=yLabel,caption=title,style='p')
        #    # error bars with points only?
        #    if len(errorBars) > 1:
        #        self.Logger.logger.logWarning("cannot plot the error bars..")
        # ------------------------------------------------------------------------

        try:
            if histo == 'OTF':
                # this step here takes a very long time!!!
                A    = self.ARRAYDATA_DATA[:,feedElement,:]
                # dimensions of the image to plot
                idim = len(A[:,0])
                jdim = len(A[0,:])
                # max and min values for the colour table
                # try to ignore first and last 20 chans when taking limits:
                try:
                    cmin = min(min(A[:,20:-20]))
                    cmax = max(max(A[:,20:-20]))
                except:
                    cmin = min(min(A))
                    cmax = max(max(A))
                # transformation matrix used to set scale for X and Y axis
                tr   = Numeric.zeros(6,'d')
                # X = TR(0) + TR(1)*I + TR(2)*J
                # Y = TR(3) + TR(4)*I + TR(5)*J
                # NOTE: the TR matrix is applied within PGPLOT: ie. in FORTRAN with
                #       array elements numbered from 1 to max. Therefore I and J in
                #       the above equation refer to these FORTRAN indices.
                # Assume always want velocity on X-axis
                # spectralAxis[i] = self._1CRVL2R + ((I-self._1CRPX2R)*self._11CD2R)
                #                 = [self._1CRVL2R - (self._1CRPX2R)*self._11CD2R] + I*self._11CD2R
                tr[0] = self._1CRVL2R - (self._1CRPX2R*self._11CD2R)
                tr[1] = self._11CD2R
                # and integration number on Y-axis
                tr[3] = 0.0
                tr[5] = 1.0
                # elements of the array to plot (all of 'em)
                # NOTE: these elements are converted in the wrapper to the FORTRAN system
                #       -> use python system starting at 0
                elX0   = 0
                elX1   = jdim-1  
                elY0   = 0
                elY1   = idim-1
                # min and max values in x and y-axis coords
                # (NOTE: I and J in FORTRAN system -> add 1 to python array elements)
                xmin = tr[0] + tr[1]*(elX0+1)
                xmax = tr[0] + tr[1]*(elX1+1)
                # the pixel used for each point goes from y-0.5 to y+0.5
                #     -> allow for this in the plot range
                ymin = tr[3] + tr[5]*(elY0+1) - 0.5
                ymax = tr[3] + tr[5]*(elY1+1) + 0.5
            else:
                xmin = min(channels)
                xmax = max(channels)
                try:
                    # remove first 10%
                    if self.SCANTYPE == 'ONOFF' or self.SCANTYPE == 'CAL':
                        tenPercent = int(0.1*len(spectrum))
                        ymin = min(spectrum[tenPercent:-tenPercent])
                        ymax = max(spectrum[tenPercent:-tenPercent])
                    # should not do it for pointings, focus or skydips!
                    else:
                        ymin = min(spectrum[:])
                        ymax = max(spectrum[:])
                except:
                    ymin = spectrum[Numeric.argmin(spectrum)]
                    ymax = spectrum[Numeric.argmax(spectrum)]


                #hha: tying to fix display problems with infinite values
                # very large and very small values are blanked 

                if(ymax>1e300 or ymax<-1e300 or ymin<-1e300 or ymin>1e300):
                    Numeric.where(abs(spectrum[:])>1.0e30,0.0,spectrum)
                    ymin=-1e20
                    ymax=1e20
                    print"hha:blanking infinite values for plot"
                    
                
                # this won't work for spectrum = 0
                if xmin == xmax:
                    xmin = channels[0]-1.0
                    xmax = channels[0]+1.0
                    self.Logger.logger.logInfo("WARNING! ppgplot: xmin=xmax")
                if ymin == ymax:
                    ymin = spectrum[0]-1.0
                    ymax = spectrum[0]+1.0
                    self.Logger.logger.logInfo("WARNING! ppgplot: ymin=ymax")


            # set the font here
            if ppgplot.numPlots >= 2:
                fontSize = 2.0
            else:
                fontSize = 1.0
            ppgplot.pgsch(fontSize)

            # the range for the plot window
            x0 = xmin-(xmax-xmin)/10.
            x1 = xmax+(xmax-xmin)/10.
            y0 = ymin-(ymax-ymin)/10.
            y1 = ymax+(ymax-ymin)/10.

            # draw the box around plot window
            ppgplot.pgenv(x0,x1,y0,y1)

            title = "%i/%i Feed: %i, %s %s, (%s, %i)"\
                    %(self.SCANNUM,self.SUBSNUM,self.FEBEPAR_USEFEED[feedElement],self.OBSTYPE,self.OBJECT,self.FEBE,self.BASEBAND)
        
            if histo == 1 and len(channels) > 1:
                ppgplot.pgbin(channels,spectrum)
            elif histo == 'OTF':
                # set the colour scale for the plot
                # (following BOA: see boa/boa/Bogli/Plot.py, setImaCol())
                #(iCiLo, iCiHi) = ppgplot.pgqcir()
                #ppgplot.pgscir(iCiLo, iCiHi)
                contrast=1
                brightness=0.5
                ncolor = 255
                ramp_intensity = Numeric.arrayrange(ncolor*1.0)/ncolor # set the color ramp
                red   = 1.0-ramp_intensity
                green = ramp_intensity
                blue  = ramp_intensity*0.0    
                ppgplot.pgctab(ramp_intensity,red,green,blue,ncolor,contrast,brightness)
                # plot the data
                ppgplot.pgimag(A,idim,jdim,elX0,elX1,elY0,elY1,cmin,cmax,tr)
                ppgplot.pgwedg_s(cmin,cmax,'RI')
            else:
                # this plots both single and multiple points
                ppgplot.pgpt(channels,spectrum,-31)
                # error bars with points only?
                if len(errorBars) > 1:
                    self.Logger.logger.logWarning("cannot plot the error bars..")
                    # there is no routine in ppgplot to plot Error bars!!
                    # therefore this does not work:
                    #yLo = spectrum-errorBars
                    #yHi = spectrum+errorBars
                    #ppgplot.pgerry(yLo,yHi,1.0)

            # Add the title and axis labels
            ppgplot.pglab(xLabel, yLabel, title)
        except:            
            ppgplot.pgenv(0,1,0,1)
            ppgplot.pgbox('BCNST',0.0,0.0,'BCNST',0.0,0.0)
            ppgplot.pglab('', '', 'Empty')

        # reset font to nominal
        ppgplot.pgsch(1)

        return


    def __reversePlanckCorrection(self, RJtemp, freq):
        """
        __reversePlanckCorrection(RJtemp, freq)
        
        Correct a Rayleigh-Jeans temperature back to real temperature
        """
        
        h = 6.626176e-34
        k = 1.380662e-23
        freq_Hz = freq*1e9
        hfk = h*freq_Hz/k

        if RJtemp != 0:
            return hfk/(Numeric.log((hfk/RJtemp) + 1.0))
        else:
            return 0.0

    # *** OPERATOR OVERLOADING METHODS ***

    def __add__(self,other):
        """
        This method overloads 'ADD' so that when it is applied to the ObsEntity
        the array of data contained within self.ARRAYDATA_DATA is added to 'other'
        The quantity that you are trying to add to the ObsEntity MUST either be another
        valid ObsEntity or a single number"""

        errorStatus = getattr(self,'errorStatus','')
        # second operand is permitted to be integer or float type
        if type(other) == float or type(other) == int:
            data1 = self.ARRAYDATA_DATA
            data2 = other
            extraErrorStatus = ''
        else:
            # Otherwise the second operand must be a valid Entity
            error = self._checkConsistent(other)
            # Raise an EntitiesError exception if not consistent
            if (error):
                raise EntitiesError(error)
            
            data1 = self.ARRAYDATA_DATA
            data2 = other.ARRAYDATA_DATA

            # combine the errorStatus which records problems in both
            # entities
            if len(errorStatus):
                errorStatus = "%s \n%s"%(errorStatus,getattr(other,'errorStatus',''))
            else:
                errorStatus = getattr(other,'errorStatus','')
        
        result = Numeric.copy.deepcopy(self)
        result.ARRAYDATA_DATA = data1 + data2

        if len(errorStatus):
            result.errorStatus = errorStatus

        return result

    def __sub__(self,other):
        """
        This method overloads 'SUBTRACT' so that when it is applied to the ObsEntity
        a second operand can be subtracted from the array of data contained within
        self.ARRAYDATA_DATA.
        The quantity that you are trying to subtract from the ObsEntity MUST either be another
        valid ObsEntity or a single number"""

        errorStatus = getattr(self,'errorStatus','')
        # second operand is permitted to be integer or float type
        if type(other) == float or type(other) == int:
            data1 = self.ARRAYDATA_DATA
            data2 = other
        else:
            # Otherwise the second operand must be a valid Entity
            error = self._checkConsistent(other)
            # Raise an EntitiesError exception if not consistent
            if (error):
                raise EntitiesError(error)
            
            data1 = self.ARRAYDATA_DATA
            data2 = other.ARRAYDATA_DATA

            # combine the errorStatus which records problems in both
            # entities
            if len(errorStatus):
                errorStatus = "%s \n%s"%(errorStatus,getattr(other,'errorStatus',''))
            else:
                errorStatus = getattr(other,'errorStatus','')
                    
        result = Numeric.copy.deepcopy(self)
        result.ARRAYDATA_DATA = data1 - data2
        if len(errorStatus):
            result.errorStatus = errorStatus

        return result

    def __rmul__(self,other):
        """
        This method overloads 'MULTIPLY' so that when it is applied to the ObsEntity
        the array of data contained within self.ARRAYDATA_DATA can be multiplied by a second
        operand.
        The quantity that you are trying to multiply the ObsEntity by MUST either be another
        valid ObsEntity or a single number"""

        errorStatus = getattr(self,'errorStatus','')
        # second operand is permitted to be integer or float type
        if type(other) == float or type(other) == int:
            data1 = self.ARRAYDATA_DATA
            data2 = other
        else:
            # Otherwise the second operand must be a valid Entity
            error = self._checkConsistent(other)
            # Raise an EntitiesError exception if not consistent
            if (error):
                raise EntitiesError(error)
            
            data1 = self.ARRAYDATA_DATA
            data2 = other.ARRAYDATA_DATA

            # combine the errorStatus which records problems in both
            # entities
            if len(errorStatus):
                errorStatus = "%s \n%s"%(errorStatus,getattr(other,'errorStatus',''))
            else:
                errorStatus = getattr(other,'errorStatus','')
                    

        result = Numeric.copy.deepcopy(self)
        result.ARRAYDATA_DATA = data1 * data2
        if len(errorStatus):
            result.errorStatus = errorStatus

        
        return result
        
    def __div__(self,other):
        """
        This method overloads 'DIVIDE' so that when it is applied to the ObsEntity
        the array of data contained within self.ARRAYDATA_DATA can be divided by a second
        operand.
        The quantity that you are trying to divide the ObsEntity by MUST either be another
        valid ObsEntity or a single number"""

        errorStatus = getattr(self,'errorStatus','')
        # second operand is permitted to be integer or float type
        if type(other) == float or type(other) == int:
            data1 = self.ARRAYDATA_DATA
            data2 = other
        else:
            # Otherwise the second operand must be a valid Entity
            error = self._checkConsistent(other)
            # Raise an EntitiesError exception if not consistent
            if (error):
                raise EntitiesError(error)
            
            data1 = self.ARRAYDATA_DATA
            data2 = other.ARRAYDATA_DATA

            # combine the errorStatus which records problems in both
            # entities
            if len(errorStatus):
                errorStatus = "%s \n%s"%(errorStatus,getattr(other,'errorStatus',''))
            else:
                errorStatus = getattr(other,'errorStatus','')
                    

        result = Numeric.copy.deepcopy(self)
        result.ARRAYDATA_DATA = data1/data2
        if len(errorStatus):
            result.errorStatus = errorStatus

        return result

    def _checkConsistent(self,other):
        """
        >> _checkConsistent(other) -> error string if not consistent
        
        This checks the current ObsEntity against another. The idea is to find
        out if the two have compatible data to be used with the algebraic methods
        __add__, __sub__, __div__ and __mul__.
        Inputs:
           other = the other 'thing' that you want to compare the Entity with.
           This can be any kind of object, number, string you like, BUT:
        Output: The output is either 0 (if the input is an acceptable match with
                                        the current ObsEntity)
                or an ErrorMessage (=string)
                This details how the input fails to meet the standards required for
                successful matching with the current ObsEntity.
            
        The following critera are checked:
        1. The 'input' must be another ObsEntity

        If it is another ObsEntity then further checks are performed:
        
        2. FEBE and BaseBand. Each ObsEntity contains a single OBSERVATION. Therefore
           it has a single FEBE/BB combination. It does not make sense to combine data
           from 2 different FEBE/BB combinations.
        3. Frequency scale. In order to keep the overloaded operators simple, the
           Frequency of each channel MUST BE IDENTICAL in both Entities.
           This is taken care of by calling self.checkChannelScale()
        4. Feeds. Also in order to keep the operators simple, the used feeds MUST be
           the same ones in both Entities. Not only the same number of Feeds but also
           EXACTLY THE SAME NAMES too.
        5. Integrations. There are 2 possibilities:
           either : There must be the same number of integrations in order to be able to make a
                    direct correspondance between the 2 entities.
           or     : One Entity must have a SINGLE integration that is applied to all
                    integrations in the other.
           If this condition isn't met there will be a 'frames not aligned' error.

        If the input passes the above conditions then the output is a logical 0"""
        
        myError = 0

        # 1st test: is it an ObsEntity?
        # if not the the following try will fail:
        try:
            # 2. FEBE/BB
            if self.FEBE != other.FEBE:
                myError = "_checkConsistent(): The FEBE must be the same in both observations "+ \
                          "FEBE of self: %s whereas FEBE of other: %s"%(self.FEBE,other.FEBE) 
                return myError
            if self.BASEBAND != other.BASEBAND:
                myError = "_checkConsistent(): The BaseBand must be the same in both observations "+ \
                          "BB of self: %i whereas BB of other: %i"%(self.BASEBAND,other.BASEBAND)
                return myError
            # 3. Frequency scale:
            check = self.checkChannelScale(other)
            if not check:
                myError = "_checkConsistent(): The Frequency/Channel scale was not the same in both "+ \
                          "observations"
                return myError
            # 4. Feeds:
            if (self.NUSEFEED != other.NUSEFEED):
                myError = EntitiesError("_checkConsistent(): Both observations must have "+ \
                                        "the same number of used feeds!")
                return myError
            else:
                for feedElement in xrange(self.NUSEFEED):
                    # I guess this could be extended so that the ORDER in which
                    # the Feeds occur need not be the same. Currently though it must be equal.
                    if self.FEBEPAR_USEFEED[feedElement] != other.FEBEPAR_USEFEED[feedElement]:
                        myError = "_checkConsistent(): The used Feeds in the two observations were not the "+ \
                                  "same. Both the Feed names and their order in the array must be identical"
                        return myError
            # 5. remaining shape: Integrations
            if Numeric.shape(self.ARRAYDATA_DATA) != Numeric.shape(other.ARRAYDATA_DATA):
                if len(self.ARRAYDATA_DATA[:,0,0]) == 1 or len(other.ARRAYDATA_DATA[:,0,0]) == 1:
                    myError = 0
                else:
                    myError = "_checkConsistent(): Both observations must either have the same number of "+ \
                              "Integrations, OR one of them must have a single Integration"
                    return myError
        except:
            myError = "_checkConsistent(): Must compare with a valid ObsEntity"
            
        return myError

    def checkChannelScale(self,other):
        """
        >> checkChannelScale(other) -> logical: true if frequency scale is identical 
        
        This checks that the channel scale are the same in two ObsEntities.
        This is neccessary because all operations are carried out assuming that corresponding
        channels in different observations have the same frequency!
        input:
             other - should be another ObsEntity
        output:
             check - logical: 0 if FALSE, 1 if TRUE
        output is only true if frequency scale is identical in both Entities.
        """

        # in order to do this, all the frequency and velocity information must be identical
        # ie. the reference channel and the frequency at reference channel
        # put it in a try loop because if the correct keywords don't exist then
        # the answer should be NO they don't match (this means other was not a valid ObsEntity)

        # start by assuming FALSE
        chanScaleCheck=[0,0,0,0,0,0]
        try:
            # if true then replaces 0 with 1:
            chanScaleCheck[0] = Numeric.equal(self.CHANNELS, other.CHANNELS)   # number channels
            chanScaleCheck[1] = Numeric.equal(self._1CRPX2F, other._1CRPX2F)  # reference channel
            chanScaleCheck[2] = Numeric.equal(self._1CRVL2F, other._1CRVL2F)  # freq at ref chan
            chanScaleCheck[3] = Numeric.equal(self._11CD2F,  other._11CD2F)   # freq resolution
            # these are strings and so must be treated in a different way:
            if self._1CUNI2F == other._1CUNI2F:
                chanScaleCheck[4] = 1  # freq unit
            if self._1SPEC2F == other._1SPEC2F:
                chanScaleCheck[5] = 1  # standard of rest
            
            if Numeric.sum(chanScaleCheck) == 6:
                check = 1
            else:
                check = 0
        except:
            check = 0

        return check
        

    # collapse/select methods:


    def sumIntegrations(self,integRange=0):
        """
        >> sumIntegrations(integRange) -> ObsEntity
        
        This method calculates the SUM of data from a range of integrations
        in the current Entity

        Input (optional):
             integRange = [minInt, maxInt] : list of 2 integers

        minInt : first Integration to include
        maxInt : last Integration to include
        
        minInt and maxInt START from 1 not zero.
        If integRange = 0 then ALL integrations will be summed (default if not specified)

        The integration time is also SUMMED over integrations.
        Feed and channel axes left unchanged.

        Times are set to the average value.
        The remaining DATAPAR parameters are collapsed using the Interpolator module.
        
        Output:
              new ObsEntity with integrations summed and updated DATAPAR.

        """
        
        NumberIntegrations = len(self.ARRAYDATA_DATA)
        result = Numeric.copy.deepcopy(self)

        # if NumberIntegrations is already 1 then don't need to do anything
        # in this case just return result which is a copy of self.
        if NumberIntegrations > 1:

            # if maxInt=0 then sum ALL integrations
            if not integRange:
                firstInt = 0
                lastInt  = NumberIntegrations
                SummedIntegrations = NumberIntegrations
            else:
                SummedIntegrations = integRange[1] - integRange[0]
                firstInt = integRange[0] - 1
                # don't subtract 1 here : python does not include the last element in the loop (ie. 1:3 is 1,2).
                lastInt  = integRange[1]
                   
            if  config.DEBUG > 1:
                self.Logger.logger.logInfo("---sumIntegrations:")        
                self.Logger.logger.logInfo("Initial number of Integrations = %i"%(NumberIntegrations))
                self.Logger.logger.logInfo("Summing Integrations from %i to %i"%(firstInt+1,lastInt+1))
                self.Logger.logger.logInfo("Final number of Integrations = 1")

            # CHECKS
            # give a warning before averaging integrations with different iswitch
            # hha: I uncommented first line as PHASE is now an integer.
            #if self.DATAPAR_PHASE*len(self.DATAPAR_PHASE) != Numeric.add.reduce(self.DATAPAR_PHASE):
            # this should work only for the old string iswitch array    
            #if self.DATAPAR_ISWITCH.count(self.DATAPAR_ISWITCH[0]) != len(self.DATAPAR_ISWITCH):

            if(self.NPHASES>1):

                # Here we will have the checks for different Phases

                if self.DATAPAR_PHASE[0]*len(self.DATAPAR_PHASE) != Numeric.add.reduce(self.DATAPAR_PHASE):
             
                    self.Logger.logger.logWarning("sumIntegrations: you are averaging different phases together")
                    self.Logger.logger.logInfo("sumIntegrations: you are averaging different phases together")

                    # log a warning message
                    if hasattr(result,'errorStatus'):
                        result.errorStatus += "\nSubscan %i: integrations with different phases were averaged!"%(self.SUBSNUM)
                    else:
                        result.errorStatus = "Subscan %i: integrations with different phases were averaged!"%(self.SUBSNUM)

            # Test if the positions of each integration are within a certain tollerance
            tollerance = 1.93e-6  # (5arcsec)
            LongVariation = max((self.DATAPAR_LONGOFF[:] - self.DATAPAR_LONGOFF[0])**2)
            LatVariation  = max((self.DATAPAR_LATOFF[:] - self.DATAPAR_LATOFF[0])**2)
            if (LongVariation > tollerance) and (LatVariation > tollerance):
                self.Logger.logger.logWarning("sumIntegrations: you are averaging different positions together")
                longDev = Numeric.sqrt(LongVariation)*3600.0
                latDev  = Numeric.sqrt(LatVariation)*3600.0
                self.Logger.logger.logInfo("LongOff max. difference: %.3f arcsec"%(longDev))
                self.Logger.logger.logInfo("LatOff max. difference: %.3f arcsec"%(latDev))
                # log a warning message
                if hasattr(result,'errorStatus'):
                    result.errorStatus = "\nSubscan %i: integrations with offsets >5 arcsec were averaged!"%(self.SUBSNUM)
                else:
                    result.errorStatus = "Subscan %i: integrations with offsets >5 arcsec were averaged!"%(self.SUBSNUM)

                # <test>
                # this was to try and remove jumps in RA and Dec that occur on a 48ms timescale in the monitor data
                #datFile = open('tmp.dat','a')
                #datFile.write(str(self.SCANNUM)+' '+str(self.SUBSNUM)+' '+str(longDev)+' '+str(latDev)+'\n')
                ##datFile.write(str(self.DATAPAR_LATOFF)+'\n')
                ##datFile.write(str(self.DATAPAR_LONGOFF)+'\n')
                ##print self.DATAPAR_LATOFF*3600#

                #mean  = Numeric.sum(self.DATAPAR_LONGOFF)/NumberIntegrations
                #stdev = Numeric.sqrt(Numeric.sum((self.DATAPAR_LONGOFF-mean)**2)/(NumberIntegrations-1.))
                #mask  = Numeric.less(Numeric.abs(self.DATAPAR_LONGOFF-mean),stdev)
                #mean2 = Numeric.sum(self.DATAPAR_LONGOFF*mask)/Numeric.sum(mask)
                #stdev2 = Numeric.sqrt(Numeric.sum(((self.DATAPAR_LONGOFF*mask)-mean2)**2)/(Numeric.sum(mask)-1.))
                #mask2  = Numeric.less(Numeric.abs(self.DATAPAR_LONGOFF-mean2),stdev2)
                #self.DATAPAR_LONGOFF[:] = Numeric.sum(self.DATAPAR_LONGOFF*mask2)/Numeric.sum(mask2)

                #mean  = Numeric.sum(self.DATAPAR_LATOFF)/NumberIntegrations
                #stdev = Numeric.sqrt(Numeric.sum((self.DATAPAR_LATOFF-mean)**2)/(NumberIntegrations-1.))
                #mask  = Numeric.less(Numeric.abs(self.DATAPAR_LATOFF-mean),stdev)
                ##rint self.DATAPAR_LATOFF*mask*3600
                #mean2 = Numeric.sum(self.DATAPAR_LATOFF*mask)/Numeric.sum(mask)
                #stdev2 = Numeric.sqrt(Numeric.sum(((self.DATAPAR_LATOFF*mask)-mean2)**2)/(Numeric.sum(mask)-1.))
                #mask2  = Numeric.less(Numeric.abs(self.DATAPAR_LATOFF-mean2),stdev2)
                ##print self.DATAPAR_LATOFF*mask2*3600
                #mean3 = Numeric.sum(self.DATAPAR_LATOFF*mask2)/Numeric.sum(mask2)
                #stdev3 = Numeric.sqrt(Numeric.sum(((self.DATAPAR_LATOFF*mask2)-mean3)**2)/(Numeric.sum(mask2)-1.))
                #mask3  = Numeric.less(Numeric.abs(self.DATAPAR_LATOFF-mean3),stdev3)
                #mean4 = Numeric.sum(self.DATAPAR_LATOFF*mask3)/Numeric.sum(mask3)
                #stdev4 = Numeric.sqrt(Numeric.sum(((self.DATAPAR_LATOFF*mask3)-mean4)**2)/(Numeric.sum(mask3)-1.))
                #mask4  = Numeric.less(Numeric.abs(self.DATAPAR_LATOFF-mean4),stdev4)
                #self.DATAPAR_LATOFF[:] = Numeric.sum(self.DATAPAR_LATOFF*mask4)/Numeric.sum(mask4)

                #print mean*3600,mean2*3600,mean3*3600,mean4*3600,stdev*3600,stdev2*3600,stdev3*3600,stdev4*3600
                
                #print DATAPAR_LATOFF*3600

                #datFile.write(str(self.DATAPAR_LATOFF)+'\n')
                #datFile.write(str(self.DATAPAR_LATOFF[0]*3600)+' '+str(self.DATAPAR_LONGOFF[0]*3600)+'\n')
                ##datFile.write(str(self.DATAPAR_LONGOFF[0])+'\n')
                #datFile.close()
                #</test>


            # *** SUM the data: ARRAYDATA_DATA and DATAPAR_INTEGTIM
            result.ARRAYDATA_DATA   = Numeric.add.reduce(self.ARRAYDATA_DATA[firstInt:lastInt,:,:])[Numeric.NewAxis,:,:]
            result.DATAPAR_INTEGTIM = Numeric.array(Numeric.add.reduce(self.DATAPAR_INTEGTIM[firstInt:lastInt]))

            # *** other attributes that change with Integration are AVERAGED using the Interpolator

            attributesToCollapse = ['DATAPAR_LST',
                                    'DATAPAR_LONGOFF',    
                                    'DATAPAR_LATOFF',    
                                    'DATAPAR_AZIMUTH', 
                                    'DATAPAR_ELEVATIO', 
                                    'DATAPAR_CBASLONG', 
                                    'DATAPAR_CBASLAT', 
                                    'DATAPAR_BASLONG', 
                                    'DATAPAR_BASLAT', 
                                    'DATAPAR_PARANGLE', 
                                    'DATAPAR_CRPIX1', 
                                    'DATAPAR_CRPIX2', 
                                    'DATAPAR_11PC', 
                                    'DATAPAR_12PC', 
                                    'DATAPAR_21PC', 
                                    'DATAPAR_22PC', 
                                    'DATAPAR_MRCVAL1', 
                                    'DATAPAR_MCRVAL2',    
                                    'DATAPAR_MLONPOLE', 
                                    'DATAPAR_MLATPOLE',
                                    'DATAPAR_DFOCUS',
                                    'DATAPAR_DPHI']   

            # add the names of the monpoint attributes:
            for attrName in vars(self).keys():
                if attrName[0:8] == 'MONPOINT':
                    attributesToCollapse.append(attrName)

            myInterpolator  = Interpolator.Interpolator()
            MJD             = getattr(self, 'DATAPAR_MJD')
            integrationTime = getattr(self, 'DATAPAR_INTEGTIM')
          
            # the Interpolator takes care of any extra dimensions that there may be in
            # dataToCollapse. (e.g. for MONITOR or DFOCUS)
            # the interpolator returns arrays
            for attribute in attributesToCollapse:
                loopLen = 1                
                dataToCollapse = getattr(self, attribute, [])
                if len(dataToCollapse):                                           
                    collapsedData, meanMJD = myInterpolator.computeCollapse(MJD[firstInt:lastInt], \
                                                                            integrationTime[firstInt:lastInt], \
                                                                            dataToCollapse[firstInt:lastInt])
                    
                    setattr(result, attribute, collapsedData)

            # Assuming that at least one of the attributes to collapse contained data
            # then meanMJD will have been filled.
            result.ARRAYDATA_MJD = meanMJD
            result.DATAPAR_MJD   = meanMJD

            # *** Other attributes changed to reflect the Collapse:
            # hha: this must be the first value


            result.DATAPAR_PHASE  =  1
            result.NINTS            = Numeric.ones(1,'i')

        return result

    def sumChannels(self,channelRange=0):
        """
        >> sumChannels(channelRange) ->  ObsEntity

        This sums over channels.
        Integration and Feed axes left unchanged.
        Can set the range with channelRange
        Default is ALL channels (when channelRange not specified)

        Input: channelRange = [minChan, maxChan] : list of 2 integers

        minChan gives the first channel to include in the sum and maxChan gives
        the last channel to INCLUDE in the sum.
        Channel numbers are taken to run from 1 to self.CHANNELS
        (ie. within the method I subtract 1 from specified values to obtain the array
        element number...which in Python starts at 0)

        Other parameters in the ObsEntity are left unchanged by this method
        (except self.CHANNELS of course)

        A new ObsEntity is returned by this method.
        """
        # this does not change the number of integrations or Feeds        
        # ChannelRange can be a list of integers

        result = Numeric.copy.deepcopy(self)
        
        try:
            if channelRange and channelRange[0] > 0 and channelRange[1] <= self.CHANNELS:
                minCh = int(channelRange[0]) - 1 
                # don't need to subtract 1 here because Python doesn't include this element in the sum:
                maxCh = int(channelRange[1])
            else:
                minCh = 0
                maxCh = self.CHANNELS
        except:
            minCh = 0
            maxCh = self.CHANNELS    
       
        result.ARRAYDATA_DATA = Numeric.add.reduce(self.ARRAYDATA_DATA[:,:,minCh:maxCh],2)[:,:,Numeric.NewAxis]
        result.CHANNELS = 1

        if config.DEBUG:
            self.Logger.logger.logInfo("Summed channels from %i to %i"%(minCh+1, maxCh))
 
        return result


    def removeAllBlanking(self, refAtt, blank=-999):
        """
        >> removeAllBlanking(refAtt, blank=-999) -> updates self
        
        This updates attributes from the ARRAYDATA and DATAPAR tables,
        removing integrations where refAtt == blank.
        Input: refAtt = attribute to examine for blanking values
                        e.g. 'DATAPAR_LONGOFF'  (string)
        Currently this attribute cannot contain strings because I use
        it in Numeric.compress and this doesn't work for lists of strings.
        (but it does update other attributes that contain strings)
        """
        refData = getattr(self,refAtt,None)

        # fix if only one row in DATAPAR
        if len(Numeric.shape(refData)) == 0:
            refData = [refData]

        if blank in refData:
            
            # extract all attribute names
            attrDic = vars(self)
            attrNames = attrDic.keys()
            
            # do a loop on all the attributes, and modify only those
            # which come from the ARRAYDATA or DATAPAR table
            for attr in attrNames:
                attrTab = string.split(attr,'_')[0]
                if attrTab in ['ARRAYDATA','DATAPAR','MONPOINT']:
                    theData = getattr(self,attr)

                    # fix if only one row
                    if len(Numeric.shape(theData)) == 0:
                        theData = [theData]

                    # if theres a problem with attributes here check the mbfits.xml matches the file.
                    #print attr,theData
		    # This may fail due to changes in the MBFITS version (e.g. from 1.56 to 1.57).
		    # In the future one may decide to implement backward compatibility by loading
		    # the corresponding XML file in the mbfits module. For now we just catch the
		    # exceptions.
		    try:
                        if type(theData[0]) != str:
                            # must specify the dimension here because "theData" can be multidimensional
                            correctedData = Numeric.compress(Numeric.not_equal(refData,blank),theData,dimension=0)
                            # print theData.shape,correctedData.shape
                            setattr(self,attr,correctedData)
                        else:
                            correctedData = []
                            for int in xrange(len(refData)):
                                if refData[int] != blank:
                                    correctedData.append(theData[int])
                            setattr(self,attr,correctedData)                        
                            # print len(theData),len(correctedData)
		    except:
		        print 'WARNING: Could not blank data in attribute %s.' % (attr)
		        print '         Probably due to MBFITS version mismatch.'
                 

            self.Logger.logger.logInfo(str(len(refData)-len(correctedData))+" integrations were removed due to blanking")
            if len(correctedData) == 0:
                msg='ALL integrations removed due to blanking!!!!'+ \
                     ' Probably means that no interpolation was performed to fill DATAPAR table - check for '+ \
                     str(blank)+' values there. This could have been caused by a miss-match in the MONITOR and '+ \
                     ' ARRAYDATA timestamps'
                raise EntitiesError(msg)
            
        #else:
            #print "0 integrations were removed due to blanking" 


    def cutIntegrations(self,integRange):
        """
        >> cutIntegrations(integRange) -> ObsEntity
        
        This returns an ObsEntity that contains only integrations
        from minInt up to and including maxInt.
        This is intended for use when e.g. you want to make a reference
        out of the first few integrations of an OTF scan

        INPUT:
             integRange = [minInt, maxInt]
        integration numbering starts at 1 (not zero)

        An updated ObsEntity is returned.
        """

        minInt = integRange[0] - 1
        maxInt = integRange[1]
        
	# Initialise the result with a copy of the input object	
        result = copy.deepcopy(self)
	# extract all attribute names
	attrDic = vars(result)
	attrNames = attrDic.keys()

	# do a loop on all the attributes, and modify only those
	# which come from the ARRAYDATA or DATAPAR table
	for attr in attrNames:
	    attrTab = string.split(attr,'_')[0]
	    if attrTab in ['ARRAYDATA','DATAPAR','MONPOINT']:
                try:
		    theData = getattr(self,attr)
                    newData = theData[minInt:maxInt]
                    setattr(result,attr,newData)
		except:
                    self.Logger.logger.logInfo("Wrong attribute type for %s" % (attr))

	# Is there any attribute outside these 2 tables that should
	# also be updated?...

        return result


    def select(self, Name, Value):
	"""
        >> select(Name, Value) -> ObsEntity
        
        Returns an ObsEntity object, containing all data
	present in the 'self' object, restricted to integrations
	where the attribute called 'Name' is equal to 'Value'.
	This makes sense only for data which change with integrations,
	which can only be in the ARRAYDATA or DATAPAR table.
        Input:
             Name = name of attribute = string
             Value = value of attribute

	"""

	# Test if the given attribute exists
	if not hasattr(self,Name):
	    raise EntitiesError("No attribute called %s in the ObsEntity"%(Name))

	# Then test if it is one of the ARRAYDATA or DATAPAR attributes
	tabName = string.split(Name,'_')[0]
        if tabName not in ['ARRAYDATA','DATAPAR','MONPOINT']:
	    raise EntitiesError("Selection can be done only on ARRAYDATA or DATAPAR attributes")

	# extract the attribute to be tested
	toTest = getattr(self,Name)

        # find the array elements where this attribute is equal to VALUE.
        goodElement = []
	for i in xrange(len(self.ARRAYDATA_DATA)):
	    if toTest[i] == Value:
                goodElement.append(i)

	# Initialise the result with a copy of the input object	
        result = Numeric.copy.deepcopy(self)

	# extract all attribute names
	attrDic = vars(result)
	attrNames = attrDic.keys()

	# do a loop on all the attributes, and modify only those
	# which come from the ARRAYDATA or DATAPAR table

	for attr in attrNames:
	    attrTab = string.split(attr,'_')[0]
	    if attrTab in ['ARRAYDATA','DATAPAR','MONPOINT']:
		theData = getattr(self,attr)
              
                # fastest way to select is to keep data in original form
                # ie. array or list.
                # it takes a very long time to transfer ARRAYDATA from a list to an array
                # if data type is string then it will be in a list not an array (e.g. ISWITCH)
                # hha: PHASE is now an array of integers
                
                if type(theData) == list or type(theData) == str:
                    newData = []
                    for i in goodElement:
                        newData.append(theData[i])                    
                else:
                    newData = Numeric.take(theData,goodElement)
                setattr(result,attr,newData)

        if config.DEBUG:
            self.Logger.logger.logInfo("Number of integrations selected is %i"%(len(goodElement)))

       
	# Is there any attribute outside these 2 tables that should
	# also be updated?...

        # If the selection was over the PHASE keyword then the number of phases in
        # the result is reduced. Therefore should update the NPHASES keyword.
        # but perhaps not..
        #if Name == 'DATAPAR_PHASE':
        #    result.NPHASES = 1
        
	return result



    def calcTRJ(self, power, frequency):
        """
        Calculate the equivalent Rayleigh-Jeans temperature for
        a certain power level
        """

        freq_Hz = frequency*1e9

        return (power * self.c**2) / (2.0 * self.k * freq_Hz**2)


    def calcBlackBodyPower(self,temperature,frequency):
        """
        Calculates the power at a certain frequency using the Planck
        formula and equivalent BlackBody temperature.
        """
        # raise error for zero division
        # must use numarray here because Error.setMode not available in Numeric
        numarray.Error.setMode(dividebyzero='raise')
        
        freq_Hz = frequency*1e9
        hfk     = self.h*freq_Hz/self.k
        upper   = (2*self.h*(freq_Hz**3))/self.c**2

        try:
            power = upper/(Numeric.exp(hfk/temperature)-1.0)
        except:
            power = temperature * 0.0

        numarray.Error.setMode(dividebyzero='warn')

        return power
    
    def calcTEBB(self,power,frequency):
        """
        Calculates the equivalent BlackBody temperature from a
        supplied power using the Planck equation.
        """
        # raise error for zero division
        # must use numarray here because Error.setMode not available in Numeric
        numarray.Error.setMode(dividebyzero='raise')
        
        freq_Hz = frequency*1e9
        hfk     = self.h*freq_Hz/self.k
        upper   = (2*self.h*(freq_Hz**3))/self.c**2

        try:
            temp = hfk / Numeric.log((upper/power) + 1.0)
        except:
            temp = power * 0.0

        numarray.Error.setMode(dividebyzero='warn')

        return temp


    def __str__(self):
	""" This method is called by the 'print' instruction. It
	defines an output format to display informations on the
	content (attributes) of an ObsEntity object. """
	# Get the list of attributes of the object
	attrDic = vars(self)		# returns a dictionnary
	attrNames = attrDic.keys()	# names of the attributes
	attrNames.sort()

	out = ''
	for attr in attrNames:
	    value = attrDic[attr]
	    if (type(value) == list):
		typ = str("list of %16s" % (type(value[0])))
		nbVal = str("%3i elements" % (len(value)))
	    elif (type(value) == type(Numeric.array([]))):
		val1D = Numeric.ravel(value)
		typ = str("array of %15s" % (type(val1D[0])))
		nbVal = str(Numeric.shape(value))
	    else:
		typ = str("%s" % (type(value)))
		nbVal = str(" = %s" % (value))

	    out = out+str("%20s %24s %20s\n" % (attr,typ,nbVal))

	return out



    def useCommandPosition(self):
        """
        use commanded source position rather than real measured position
        This replaces the current LONGOFF and LATOFF with commanded position
        - to get around problems caused by ra dec jumps in the monitor points.

        Only valid for EQEQEQ
        """
        if self.USRFRAME == 'EQEQEQ':

            if self.MOVEFRAM == 'T':
                crval1 = self.DATAPAR_MCRVAL1
                crval2 = self.DATAPAR_MCRVAL2
            else:
                crval1 = self.CRVAL1
                crval2 = self.CRVAL2

            self.DATAPAR_LONGOFF = (self.DATAPAR_CBASLONG - crval1) * Numeric.cos(crval2/180.*Numeric.pi)
            self.DATAPAR_LATOFF  = self.DATAPAR_CBASLAT - crval2

            self.Logger.logger.logWarning("Using COMMANDED positions to calculate offsets!!")

        else:
            self.Logger.logger.logInfo("Could not use commanded position for offsets: only for EQEQEQ")

            

class EntitiesError(Exception):

    """ A class used to generate exceptions related
    to the Entities.py module. """
    
    def __init__(self, msg, value=None):
        self.value = value
        self.msg = msg

    def __str__(self):
        return `self.msg`



























